# Produced for the LABOCA project
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#

"""
NAM: BoaFlagHandler.py (file)
DES: Contains Class FlagHandler for bitwise manipulation of flags.
"""                    
# ---------------------------------------------------------------------
# ---- Import ---------------------------------------------------------
# ---------------------------------------------------------------------

from Numeric import *

# ---------------------------------------------------------------------
# ---- Class FlagHandler ----------------------------------------------
# ---------------------------------------------------------------------

class FlagHandler:
    """
    NAM: FlagHandler (Class)
    DES: Provides methods to manipulate and query flag arrays bitwise.
         Here, a flag array is a Numeric.array, each element of which
         represents n independent flags, with n the number of bits of
         each array element. The n independent flags are enumerated with
         the flag value, ranging from 1 (the rightmost bit) to n (the
         leftmost bit).

         Throughout the class, the nomenclature is as follows:

         - aFlags (Numeric.array):
             The flag array to be manipulated or queried.

             The flag array can be of any integer type, although UInt8,
             UInt16, and UInt32 are the most natural choices.

             The shape of the array is arbitrary.

         - iFlags (integer, list of integers, or empty list):
             The flag values used for manipulating or queriing the flag
             array.

             Valid flag values are from 1 to n, with n the number of bits
             of each element of aFlags.

             If iFlags is the empty list, the list [1,2,...,n] is assumed.
    """

    # -----------------------------------------------------------------

    def _getBitmask(self, aFlags, iFlags):
        # DES: Determine the bitmask corresponding to iFlags.
        # INP: aFlags (Numeric.array)             : Flag array
        #      iFlags (int, int list, empty list) : Flag values
        #      (see also doc string of class FlagHandler)
        # OUT: Numeric.array of the same typecode as aFlags,
        #      shape(1,), and value such that bits corresponding
        #      to iFlags are 1, all others 0
        typecode = aFlags.typecode()
        validFlagValues = self.getValidFlagValues(aFlags)
        
        zero = zeros(1,typecode=typecode)
        one = ones(1,typecode=typecode)

        bitmask = zero

        if iFlags==[]:
            iFlags = validFlagValues
        if type(iFlags)==type(1):
            iFlags = [iFlags]

        for iFlag in iFlags:
            if not iFlag in validFlagValues:
                raise "Invalid flag"
            bitmask |= (one<<(iFlag-1)).astype(typecode)

        return bitmask

    # -----------------------------------------------------------------

    def getValidFlagValues(self, aFlags):
        """
        NAM: FlagHandler.getValidFlagValues (Method)
        DES: Returns a list of all valid flag values for
             flag array aFlags
        INP: aFlags (Numeric.array)             : Flag array
             (see also doc string of class FlagHandler)
        OUT: (int list) : All valid flag values for aFlags
        """
        return range(1, aFlags.itemsize()*8+1)

    # -----------------------------------------------------------------

    def setAll(self, aFlags, iFlags=[]):
        """
        NAM: FlagHandler.setAll (Method)
        DES: Sets the flag values iFlags for all elements of
             flag array aFlags
        INP: aFlags (Numeric.array)             : Flag array
             iFlags (int, int list, empty list) : Flag values
             (see also doc string of class FlagHandler)
        """
        bitmask = self._getBitmask(aFlags, iFlags)
        bitwise_or(aFlags, bitmask, aFlags)

    # -----------------------------------------------------------------

    def unsetAll(self, aFlags, iFlags=[]):
        """
        NAM: FlagHandler.unsetAll (Method)
        DES: Unsets the flag values iFlags for all elements of
             flag array aFlags
        INP: aFlags (Numeric.array)             : Flag array
             iFlags (int, int list, empty list) : Flag values
             (see also doc string of class FlagHandler)
        """
        bitmask = self._getBitmask(aFlags, iFlags)
        bitwise_and(aFlags, ~bitmask, aFlags)

    # -----------------------------------------------------------------

    def setOnIndex(self, aFlags, index, iFlags=[]):
        """
        NAM: FlagHandler.setOnIndex (Method)
        DES: Sets the flag values iFlags for a single element of
             flag array aFlags
        INP: aFlags (Numeric.array)             : Flag array
             iFlags (int, int list, empty list) : Flag values
             (see also doc string of class FlagHandler)
             index (int, int list) : Index of the element of aFlags; must
                                     specify a single element.
        """
        bitmask = self._getBitmask(aFlags, iFlags)

        if type(index)==type(1):
            index = [index]
        if len(index)!=len(aFlags.shape):
            raise "Invalid index"

        aFlags[index] =  \
             (aFlags[index] | bitmask).astype(aFlags.typecode())[0]


    # -----------------------------------------------------------------

    def unsetOnIndex(self, aFlags, index, iFlags=[]):
        """
        NAM: FlagHandler.unsetOnIndex (Method)
        DES: Unsets the flag values iFlags for a single element of
             flag array aFlags
        INP: aFlags (Numeric.array)             : Flag array
             iFlags (int, int list, empty list) : Flag values
             (see also doc string of class FlagHandler)
             index (int, int list) : Index of the element of aFlags; must
                                     specify a single element.
        """
        bitmask = self._getBitmask(aFlags, iFlags)
        
        if type(index)==type(1):
            index = [index]
        if len(index)!=len(aFlags.shape):
            raise "Invalid index"

        aFlags[index] =  \
            (aFlags[index] & ~bitmask).astype(aFlags.typecode())[0]

    # -----------------------------------------------------------------

    def setOnMask(self, aFlags, aMask, iFlags=[]):
        """
        NAM: FlagHandler.setOnMask (Method)
        DES: Sets the flag values iFlags for all elements of
             flag array aFlags specified by aMask
        INP: aFlags (Numeric.array)             : Flag array
             iFlags (int, int list, empty list) : Flag values
             (see also doc string of class FlagHandler)
             aMask (Numeric.array) : Mask specifiing the elements of
                                     aFlags to be manipulated.
                                     The shape of aMask must be the shape
                                     of aFlags.
                                     Only elements of aFlags, for which aMask
                                     is not 0, will be manipulated.
        """
        mask = aMask.astype(aFlags.typecode())
        if mask.shape!=aFlags.shape:
            raise "Invalid mask"

        bitmask = self._getBitmask(aFlags, iFlags)

        putmask(mask, mask, bitmask)
        bitwise_or(aFlags, mask, aFlags)

    # -----------------------------------------------------------------

    def unsetOnMask(self, aFlags, aMask, iFlags=[]):
        """
        NAM: FlagHandler.unsetOnMask (Method)
        DES: Unsets the flag values iFlags for all elements of
             flag array aFlags specified by aMask
        INP: aFlags (Numeric.array)             : Flag array
             iFlags (int, int list, empty list) : Flag values
             (see also doc string of class FlagHandler)
             aMask (Numeric.array) : Mask specifiing the elements of
                                     aFlags to be manipulated.
                                     The shape of aMask must be the shape
                                     of aFlags.
                                     Only elements of aFlags, for which aMask
                                     is not 0, will be manipulated.
        """
        mask = aMask.astype(aFlags.typecode())
        if mask.shape!=aFlags.shape:
            raise "Invalid mask"

        bitmask = self._getBitmask(aFlags, iFlags)

        putmask(mask, mask, bitmask)
        bitwise_and(aFlags, ~mask, aFlags)


    # -----------------------------------------------------------------

    def isSetMask(self, aFlags, iFlags=[]):
        """
        NAM: FlagHandler.isSetMask (Method)
        DES: Returns a mask that indicates for which elements
             of aFlags at least one flag value specified by
             iFlags is set.
        INP: aFlags (Numeric.array)             : Flag array
             iFlags (int, int list, empty list) : Flag values
             (see also doc string of class FlagHandler)
        OUT: (Numeric.array) : Mask indicating that at least one flag value
                               specified by iFlags is set for the corresponding
                               element of aFlags.
        """
#        return where(aFlags & bitmask, 1, 0)
# The following is faster by a factor ~2
        bitmask = self._getBitmask(aFlags, iFlags)
        return logical_and(aFlags & bitmask, ones(1,typecode=aFlags.typecode()))

    # -----------------------------------------------------------------

    def isUnsetMask(self, aFlags, iFlags=[]):
        """
        NAM: FlagHandler.isUnsetMask (Method)
        DES: Returns a mask that indicates for which elements
             of aFlags none of the flag values specified by
             iFlags is set.
        INP: aFlags (Numeric.array)             : Flag array
             iFlags (int, int list, empty list) : Flag values
             (see also doc string of class FlagHandler)
        OUT: (Numeric.array) : Mask indicating that none of the flag values
                               specified by iFlags is set for the corresponding
                               element of aFlags.
        """
#        return logical_not(self.isSetMask(aFlags, iFlags))
# The following is faster by a factor ~2
        bitmask = self._getBitmask(aFlags, iFlags)
        return logical_not(aFlags & bitmask)

    # -----------------------------------------------------------------

    def nSet(self, aFlags, iFlags=[]):
        """
        NAM: FlagHandler.nSet (Method)
        DES: Returns the number of elements of aFlags for which
             at least one flag value specified by iFlags is set.
        INP: aFlags (Numeric.array)             : Flag array
             iFlags (int, int list, empty list) : Flag values
             (see also doc string of class FlagHandler)
        OUT: (int) : Number of elements of aFlags for which at least
                     one flag value specified by iFlags is set.
        """
        return len(nonzero(self.isSetMask(aFlags, iFlags).flat))

    # -----------------------------------------------------------------

    def nUnset(self, aFlags, iFlags=[]):
        """
        NAM: FlagHandler.nUnset (Method)
        DES: Returns the number of elements of aFlags for which
             none of the flag values specified by iFlags is set.
        INP: aFlags (Numeric.array)             : Flag array
             iFlags (int, int list, empty list) : Flag values
             (see also doc string of class FlagHandler)
        OUT: (int) : Number of elements of aFlags for which none
                     of the flag values specified by iFlags is set.
        """
        return len(nonzero(self.isUnsetMask(aFlags, iFlags).flat))

    # -----------------------------------------------------------------

    def isSetOnIndex(self, aFlags, index, iFlags=[]):
        """
        NAM: FlagHandler.isSetOnIndex (Method)
        DES: Returns 1 if at least one flag value specified by iFlags
             is set for a single element of flag array aFlags.
        INP: aFlags (Numeric.array)             : Flag array
             iFlags (int, int list, empty list) : Flag values
             (see also doc string of class FlagHandler)
             index (int, int list) : Index of the element of aFlags; must
                                     specify a single element.
        OUT: (int) : 1 if at least one flag value specified by iFlags
                     is set, 0 else.
        """
        bitmask = self._getBitmask(aFlags, iFlags)
        
        if type(index)==type(1):
            index = [index]
        if len(index)!=len(aFlags.shape):
            raise "Invalid index"

        if (aFlags[index] & bitmask):
            isSet = 1
        else:
            isSet = 0
        return isSet

    # -----------------------------------------------------------------

    def isUnsetOnIndex(self, aFlags, index, iFlags=[]):
        """
        NAM: FlagHandler.isSetOnIndex (Method)
        DES: Returns 1 if none of the flag values specified by iFlags
             are set for a single element of flag array aFlags.
        INP: aFlags (Numeric.array)             : Flag array
             iFlags (int, int list, empty list) : Flag values
             (see also doc string of class FlagHandler)
             index (int, int list) : Index of the element of aFlags; must
                                     specify a single element.
        OUT: (int) : 1 if none of the flag values specified by iFlags
                     is set, 0 else.
        """
        return logical_not(self.isSetOnIndex(aFlags, index, iFlags))
