# Copyright (C) 2002-2006
# Max-Planck-Institut fuer Radioastronomie Bonn
# Argelander Institut fuer Astronomie
# Astronomisches Institut der Ruhr-Universitaet Bochum
#
# Produced for the LABOCA project
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#

"""
NAM: BoaFocus.py (module)
DES: contains the BoA focus class
"""
__version__=  '$Revision: 2096 $'
__date__=     '$Date: 2007-05-23 02:36:24 +0200 (Mi, 23 Mai 2007) $'

# ---------------------------------------------------------------------
# ---- Import ---------------------------------------------------------
# ---------------------------------------------------------------------
from Numeric import *
from boa           import BoaPointing, BoaConfig
from boa.Bogli     import Plot, MultiPlot
from boa.Utilities import fitParabola, modelparabola
from boa.fortran   import fStat

# ---------------------------------------------------------------------
# ---- Boa Focus Class ------------------------------------------------
# ---------------------------------------------------------------------

class Focus(BoaPointing.Point):
  """
  NAM: Focus (class)
  DES: An object of this class is responsible for the focus reduction
       of single or multiple scans and provides the offsets.
  """

  def __init__(self):
    """
    DES: Initialise an instance
    """
    BoaPointing.Point.__init__(self)
    self.ResultOffsets = []
    self.ResultFluxes = []

  # -------------------------------------------------------------------
  # ---- public methods -----------------------------------------------
  # -------------------------------------------------------------------
  def solveFocus(self):
    """
    DES: compute the optimal focus position
    """

    refChan = self.BolometerArray.RefChannel
    modType = self.ScanParam.ScanType
    nSubscan = len(self.ScanParam.SubscanNum)  # subscans read in

    iflux  = zeros((nSubscan),Float32)
    isdev  = zeros((nSubscan),Float32)
    ifocus = zeros((nSubscan),Float32)

    for i in range(nSubscan):
      flux    = self.getChanData('flux',refChan,subscans=[self.ScanParam.SubscanNum[i]])
      focus   = self.getChanData(modType,refChan,subscans=[self.ScanParam.SubscanNum[i]])
      iflux[i] = fStat.f_median(flux)
      isdev[i] = fStat.f_rms(flux,iflux[i])
      ifocus[i] = fStat.f_median(focus)
      
    xLabel = modType +' Offsets [mm] '
    yLabel = 'flux density [arb u.] Channel :'+str(refChan)

    result = fitParabola(ifocus,iflux,isdev)

    x = arange(101.)/100.*(max(ifocus)-min(ifocus))+min(ifocus)
    y = modelparabola(result.params,x)
    
    Plot.plot(ifocus,iflux,style='p',\
              labelX=xLabel,labelY=yLabel,\
              caption=self.ScanParam.caption())
    # TODO : need plot with error bar.
    #Plot.plot(ifocus,iflux,style='p',ci=1,overplot=1)
    Plot.plot(x,y,overplot=1,style='l')
    print "Offset : %f5.2 +- %f5.2" \
          %(result.params[1]/(-2.0*result.params[2]),result.perror[1]/(-2.0*result.params[2]))

  def reduce(self,datasetName='',obstoProc=[]):
    """
    DES: Process a Focus scan - this method is called by the apexCalibrator
    INP: (string) datasetName: path to the dataset to be reduced
         (i list)   obstoProc: list of subscans to consider (default: all)
    """
    self.read(inFile=datasetName,subscans=obstoProc)
    self.FF_Median = self.BolometerArray.Gain
    refChan = self.BolometerArray.RefChannel
    # TODO: medianbase or zeroStart, but NOT by subscan
    self.medianNoiseRemoval(computeFF=0,nbloop=2)
            
    refChan = self.BolometerArray.RefChannel
    flux = self.getChanData('flux',refChan)

    modType = self.ScanParam.ScanType
    focus = self.getChanData(modType,refChan)

    resultOff = []
    resultFlux = []
    for obs in obstoProc:
      flux    = self.getChanData('flux',refChan,subscans=[obs])
      focus   = self.getChanData(modType,refChan,subscans=[obs])
      resultOff.append(fStat.f_median(focus))
      resultFlux.append(fStat.f_median(flux))
    
    self.ResultOffsets = resultOff
    self.ResultFluxes = resultFlux

   
