"""
NAM: BoaMapping.py (module)
DES: contains the BoAMapping and Image classes
"""
__version__=  '$Revision: 2355 $'
__date__=     '$Date: 2007-08-21 01:50:31 +0200 (Di, 21 Aug 2007) $'

#----------------------------------------------------------------------------------
#----- Import ---------------------------------------------------------------------
#----------------------------------------------------------------------------------
from Numeric import *
from LinearAlgebra import *
import cPickle,string

from boa           import BoaDataAnalyser, BoaMessageHandler, BoaFits
from boa           import BoaConfig, BoaError
from boa.Bogli     import Plot, MultiPlot
from boa.fortran   import fUtilities, fMap, fStat
from boa.Utilities import Timing, compressNan,tolist_boa
from boa.Utilities import modelBaseEllipticalGaussian,fitBaseEllipticalGaussian

#TODO ppgplot should not be called directly
from ppgplot import *  # for getPixel

import os.path

#----------------------------------------------------------------------------------
#----- Image Class ----------------------------------------------------------------
#----------------------------------------------------------------------------------

class Image:
    """
    NAM: Image (class)
    DES: An object of this class describes an image and its axis
    """


    # TODO add method getPixel putPixel and/or getRegion setRegion
    def __init__(self):

        self.__MessHand=BoaMessageHandler.MessHand(self.__module__)
        
        self.Header = {
            'Object':   'Unknown',\
            'Telescope':  'None',\
            'FeBe':       'None'\
            }

        self.Data     = []   # The bitmap map (2D arrays)
	self.Weight   = []   # The weights per pixel
	self.Coverage = []   # The coverage per pixel, i.e. weight=1
        self.RmsBeam  = 0.   # rms/beam in current unit
	self.BeamSize = 0.   # beam size, in same unit as WCS
        
        self.WCS = {
            # Number of Axis (2 for simple images)
            'NAXIS': 2,\
            \
            # time of observation
            'MJD-OBS': 'None',\
            # frame of reference
            'RADESYS':'FK5',\
            # coordinate epoch
            'EQUINOX': 2000.0,\
            \
            # #of pixel
            'NAXIS1': 0,\
            # Type 
            'CTYPE1': 'X',\
            # pixel scale
            'CDELT1': 1.0,\
            # unit
            'CUNIT1': 'dummy',\
            # Pixel coordinate of reference point
            'CRPIX1': 0.0,\
            # celestial coordinate of reference point
            'CRVAL1': 0.0,\
            \
            'NAXIS2': 0,\
            'CTYPE2': 'Y',\
            'CDELT2': 1.0,\
            'CUNIT2': 'dummy',\
            'CRPIX2': 0.0,\
            'CRVAL2': 0.0,\
            \
            # linear transformation matrix
            'PC1_1': 1,\
            'PC1_2': 0,\
            'PC2_1': 0,\
            'PC2_2': 1}

    def __str__(self):
        """
        DES:Defines a string, shown when the print instruction is used.
        """
        out = "Image of %s with %s on %s (%ix%i pixels)"%\
              (self.Header['Object'],\
               self.Header['FeBe'],\
               self.Header['Telescope'],\
               self.WCS['NAXIS1'],self.WCS['NAXIS2'])

        return out
            
    def wcs2pix(self,X,Y):
        """
        DES: Convert from physical coordinates described by self.WCS
             to pixel coordinates
        INP: float (X,Y) : the physical coordinates to convert from
        OUT: float (i,j) : the pixel coordinates

	     We should switch to libwcs at some point
        """
        if self.WCS['CUNIT1'] == 'arcsec':
            cdeltUnit = 1./3600.
        else:
            cdeltUnit = 1.
        AXIS1 = array([self.WCS['NAXIS1'],self.WCS['CRPIX1'],self.WCS['CDELT1'],
                       self.WCS['CRVAL1'],cdeltUnit])
        AXIS2 = array([self.WCS['NAXIS2'],self.WCS['CRPIX2'],self.WCS['CDELT2'],
                       self.WCS['CRVAL2'],cdeltUnit])
        i,j = fMap.wcs2pix(X,Y,AXIS1,AXIS2)
        if not rank(X) and not rank(Y):
            i = i[0]
            j = j[0]
        return i,j
    
    def wcs2phy(self,i,j):
        """
        DES: Convert from pixel coordinates to physical (world) coordinates
        INP: float (i,j) : the pixel coordinates to convert from
        OUT: float (X,Y) : the physical coordinates

	     We should switch to libwcs at some point
        """
        if self.WCS['CUNIT1'] == 'arcsec':
            cdeltUnit = 1./3600.
        else:
            cdeltUnit = 1.
	AXIS1 = array([self.WCS['NAXIS1'],self.WCS['CRPIX1'],self.WCS['CDELT1'],
                       self.WCS['CRVAL1'],cdeltUnit])
	AXIS2 = array([self.WCS['NAXIS2'],self.WCS['CRPIX2'],self.WCS['CDELT2'],
                       self.WCS['CRVAL2'],cdeltUnit])
        X,Y = fMap.wcs2phy(i,j,AXIS1,AXIS2)
        if not rank(i) and not rank(j):
            X = X[0]
            Y = Y[0]
        return X,Y
    
   
    def computeWCS(self,pixelSize,sizeX=[],sizeY=[],minmax=[]):
        """
        DES: fill main WCS keywords according to pixel size and map limits
        INP: (int) pixelSize = size of pixel in acrsecond
             (float) sizeX = map limits in azimuth, in arcsecond
             (float) sizeY = map limits in elevation, in arcsecond
             (float) minmax = [minAzoff,maxAzoff,minEloff,maxEloff] in this order
        """

        # TODO add shift for 'CRPIX1/2' or 'CRVAL1/2' to be general and allow for
        # offset in  maps

        # determine coordinate limits and map size
        if sizeX == []:
            minAzoff = minmax[0]
            maxAzoff = minmax[1]
        else:
            minAzoff = sizeX[0]
            maxAzoff = sizeX[1]
            
        if sizeY == []:
            minEloff = minmax[2]
            maxEloff = minmax[3]
        else:
            minEloff = sizeY[0]
            maxEloff = sizeY[1]
           
        Xcenter = (minAzoff + maxAzoff)/2.
        Ycenter = (minEloff + maxEloff)/2.
        if self.WCS['CUNIT1'] == 'arcsec':
            cosY = cos(Ycenter*pi/180./3600.)
        else:
            cosY = cos(Ycenter*pi/180.)
            
        dimX = int(ceil(abs(maxAzoff - minAzoff)/pixelSize*cosY +1.))
        dimY = int(ceil(abs(maxEloff - minEloff)/pixelSize +1.))
        
        # Size of the image
        self.WCS['NAXIS1'] = dimX
        self.WCS['NAXIS2'] = dimY
        
        # Size of pixels
        self.WCS['CDELT1'] = (maxAzoff - minAzoff)/abs(maxAzoff - minAzoff)*pixelSize
        self.WCS['CDELT2'] = (maxEloff - minEloff)/abs(maxEloff - minEloff)*pixelSize
        
        # put the reference in the center of the image ...
        self.WCS['CRPIX1'] = (dimX-1)/2.
        self.WCS['CRPIX2'] = (dimY-1)/2.
        
        # ... corresponding to the center of the map
        self.WCS['CRVAL1'] = Xcenter
        self.WCS['CRVAL2'] = Ycenter

    def physicalCoordinates(self):
        """
        DES: return arrays with physical units corresponding to the map
        """
        allAz = self.wcs2phy(arange(self.WCS['NAXIS1']),repeat([0],self.WCS['NAXIS1']))[0]
        allEl = self.wcs2phy(repeat([0],self.WCS['NAXIS2']),arange(self.WCS['NAXIS2']))[1]
        repeatAz = repeat(allAz,self.WCS['NAXIS2'])
        resultAz = reshape(repeatAz,(self.WCS['NAXIS1'],self.WCS['NAXIS2']))
        repeatEl = repeat(allEl,self.WCS['NAXIS1'])
        resultEl = transpose(reshape(repeatEl,(self.WCS['NAXIS2'],self.WCS['NAXIS1'])))
        return (resultAz,resultEl)


    def display(self, weight=0, coverage=0, style='idl4', caption='', wedge=1, \
		aspect=1, overplot=0, doContour=0, levels=[], labelContour=0,\
		limitsX = [], limitsY = [], limitsZ = [], \
                showRms=0, rmsKappa=2 ):
        """
        DES: show the reconstructed maps in (Az,El)
        INP: (boolean) weigth,coverage : plot the rms or weight map instead of signal map
             (string)  style           : the style used for the color (default idl4)

             (string)  caption       : the caption of the plot (default '')
             (flt array) limitsX/Y/Z : the limits in X/Y/intensity
             (boolean) wedge         : draw a wedge ? (default : yes)
             (boolean) aspect        : keep the aspect ratio (default : yes)
             (boolean) overplot      : should we overplot this image (default : no)
             (boolean) doContour     : draw contour instead of map (default : no)
             (float array) levels    : the levels of the contours
                                       (default : intensity progression)
             (boolean) labelContour  : label the contour (default : no)
             (boolean) showRms       : compute and display rms/beam? (def: yes)
        """

	toPlot = self.Data
        WCS    = self.WCS

	if weight:
            toPlot = self.Weight
	if coverage:
            toPlot = self.Coverage

        labelX=''
        if WCS['CTYPE1'][0:4] == 'OLON':
            labelX = "\gD Az"
        if WCS['CTYPE1'][0:4] == 'RA--':
            labelX="\ga"
            if WCS['EQUINOX'] == 2000.0:
                labelX += ' (J2000)'
            if WCS['EQUINOX'] == 1950.0:
                labelX += ' (B1950)'
        labelX += " ["+WCS['CUNIT1']+"]"

        labelY=''
        if WCS['CTYPE2'][0:4] == 'OLAT':
            labelY="\gD El"
        if WCS['CTYPE2'][0:4] == 'DEC-':
            labelY="\gd"
            if WCS['EQUINOX'] == 2000.0:
                labelY+=' (J2000)'
            if WCS['EQUINOX'] == 1950.0:
                labelY+=' (B1950)'
        labelY += " ["+WCS['CUNIT2']+"]"

        Plot.draw(toPlot,wedge=wedge, WCS=WCS,\
		  labelX=labelX, labelY=labelY, caption=caption, \
		  limitsX = limitsX, limitsY = limitsY, limitsZ = limitsZ,\
		  nan=1, style=style,aspect=aspect, overplot=overplot,\
		  doContour=doContour,levels = levels, labelContour=labelContour)

        if showRms:
            # compute rms/beam
            Xrange, Yrange = [],[]
            if limitsX:
                x1,y1 = self.wcs2pix(limitsX[0],WCS['CRVAL2'])
                x2,y2 = self.wcs2pix(limitsX[1],WCS['CRVAL2'])
                Xrange = [int(x1)+1,int(x2)]
            if limitsY:
                x1,y1 = self.wcs2pix(WCS['CRVAL1'],limitsY[0])
                x2,y2 = self.wcs2pix(WCS['CRVAL1'],limitsY[1])
                Yrange = [int(y1),int(y2)+1]

            self.computeRms(rmsKappa=rmsKappa,
                            limitsX=Xrange,limitsY=Yrange)
        
    def smoothBy(self,Size):
       """
       DES: Smooth the image with a 2D Gaussian of given FWHM.
            Smoothing is peak-normalised, therefore conserves Jy/beam
            as unit.
       INP: (float) Size : the FWHM of the smoothing gaussian
       """
       
       # Compute a kernel of the same pixelSize as the image
       pixsize = abs(self.WCS['CDELT2'])
       smoothingKernel = Kernel(pixsize,Size)
       self.smoothWith(smoothingKernel)
       # Normalisation of the data plane: pixel unit is flux/beam
       # so, update to the new beam size
       newbeam = sqrt(self.BeamSize**2 + Size**2)
       self.Data   *= array((newbeam**2)/ (self.BeamSize**2),'f')
       self.Weight *= array((newbeam**2)/ (self.BeamSize**2),'f')       
       self.BeamSize = newbeam

    def smoothWith(self,kernel):
       """
       DES: smooth the image with the given kernel
       INP: (kernel) : the kernel
       """
       
       if abs(self.WCS['CDELT1']) == abs(kernel.WCS['CDELT1']) and \
              abs(self.WCS['CDELT2']) == abs(kernel.WCS['CDELT2']):
           self.Data     = fMap.ksmooth(self.Data,     kernel.Data)
           self.Weight   = fMap.ksmooth(self.Weight,   kernel.Data)
           self.Coverage = fMap.ksmooth(self.Coverage, kernel.Data)

    def getPixel(self,nbPix=3):
        """
        DES: allow user to get pixel values using mouse
        INP: (int)  nbPix : size of area to compute average (default 3x3)
        """

        self.__MessHand.info("Click left to get one pixel, mid to get average over " +\
                           str(nbPix*nbPix)+", right to exit (on Data array only)")

        data = self.Data
        WCS  = self.WCS
        
        x,y = 0,0
        char = ''
        while(char != 'X'):
            if (char == 'D'):
                #average over [x1,x2] x [y1,y2]
                offset = (nbPix-1)/2.
                if offset <= i < WCS['NAXIS1']-offset and \
                       offset <= j < WCS['NAXIS2']-offset:
                    x1 = int(i-offset)
                    x2 = int(i+offset+1)
                    y1 = int(j-offset)
                    y2 = int(j+offset+1)
                    
                    listVal = self.Data[x1:x2,y1:y2]
                    listVal = ravel(listVal)
                    
                    listVal, nNan = compressNan([listVal])
                    if len(listVal) > 0:
                        val = sum(listVal)/float(len(listVal))
                        self.__MessHand.info("[%ix%i]-%i: (%6.2f,%6.2f) = %6.2f" \
                                             %(nbPix,nbPix,len(listVal),x,y,val))
                else:
                    self.__MessHand.warning("some pixel(s) are outside the plot")
            elif (char == 'A'):
                # Print a single pixel 
                if 0 <= i < WCS['NAXIS1'] and 0 <= j < WCS['NAXIS2']:
                    val = self.Data[i,j]
                    self.__MessHand.info("("+str("%7.3f" % x)+","+str("%7.3f" % y)+") = "+str(val))
                else:
                    self.__MessHand.warning("pixel outside the plot")

            x,y,char = pgband(7,0,x,y)
            i,j = self.wcs2pix(x,y)
            i = int(i)
            j = int(j)

    #--------------------------------------------------------------------------------
    def zoom(self,mouse=1,style='idl4',wedge=1,\
                limitsZ=[],aspect=1,limitsX=[],limitsY=[],caption=None,\
                doContour=0,levels=[],showRms=1,rmsKappa=2):
        """
        DES: allow the user to select a region in the map to zoom in
        INP: (bool) mouse: use the mouse? (default: yes)
               (other parameters: same as display)
        """

        if not self.Data:
            self.__MessHand.error('No map computed yet')
            return

        if not mouse:
            if not limitsX or not limitsY:
                self.MessHand.error('X/Y limits must be given if not using the mouse')
                return
            else:
                limX = limitsX
                limY = limitsY
        else:
            self.__MessHand.info('Use the mouse cursor to select 2 opposite corners')
            WCS  = self.WCS        
            x,y = 0,0
            char = ''
            while char == '':
                x,y,char = pgband(7,0,x,y)
                i,j = self.wcs2pix(x,y)
                if not(0 <= i < WCS['NAXIS1'] and 0 <= j < WCS['NAXIS2']):
                    self.__MessHand.warning("pixel outside the plot")
                    char = ''
            x1,y1 = x,y
            self.__MessHand.longinfo("Corner1: ("+str("%8.4f" % x)+","+str("%8.4f" % y)+")")
            char = ''
            while char == '':
                x,y,char = pgband(7,0,x,y)
                i,j = self.wcs2pix(x,y)
                if not(0 <= i < WCS['NAXIS1'] and 0 <= j < WCS['NAXIS2']):
                    self.MessHand.warning("pixel outside the plot")
                    char = ''
            x2,y2 = x,y
            self.__MessHand.longinfo("Corner2: ("+str("%8.4f" % x)+","+str("%8.4f" % y)+")")

            # sort limits
            limX = [max([x1,x2]),min([x1,x2])]
            limY = [min([y1,y2]),max([y1,y2])]
        if not caption:
            caption = 'Zoom'
        self.__MessHand.info('limitsX=['+str("%8.4f" % max(limX))+','+ \
                             str("%8.4f" % min(limX)+'],')+ \
                             'limitsY=['+str("%8.4f" % min(limY))+','+ \
                             str("%8.4f" % max(limY)+'] '))
        self.display(style=style,
                     limitsX=limX,limitsY=limY,limitsZ=limitsZ,
                     aspect=aspect,caption=caption,wedge=wedge,
                     doContour=doContour,levels=levels,
                     showRms=showRms, rmsKappa=rmsKappa)
            

    #--------------------------------------------------------------------------------
    def extractSource(self,gradient=0, circular=0, radius=-10, \
                      Xpos = 0., Ypos = 0., fixedPos = 0):
        """
        DES: fit a 2D Gaussian on  a map -
        """
        if not self.Data:
            self.__MessHand.error('No map computed yet')
            return
        
        fwhm = self.BeamSize  # that's in the map coordinates unit
        # If map in EQ system, use central position as starting point
        if self.WCS['CTYPE1'].find('RA') >= 0 or self.WCS['CTYPE2'].find('DEC') >= 0:
            if not Xpos:  # but only if not provided by the user
                Xref = self.WCS['CRVAL1']
            else:
                Xref = Xpos
            if not Ypos:
                Yref = self.WCS['CRVAL2']
            else:
                Yref = Ypos
        else:
            Xref, Yref = Xpos, Ypos
        
        mapArray          = self.Data
        weightArray       = self.Weight
        azArray, elArray  = self.physicalCoordinates()
        
        # put everything into 1D easier for radius compression
        mapArray    = ravel(mapArray)
        weightArray = ravel(weightArray)
        azArray     = ravel(azArray)
        elArray     = ravel(elArray)

        if radius < 0:
            radius = abs(radius)*fwhm

        # Select data to a certain radius from the given X/Y pos
        if radius != 0:
            good_for_fit = less(sqrt((azArray-Xref)**2+(elArray-Yref)**2),radius)
            azArray,n   = fUtilities.compress(azArray,good_for_fit,1)
            azArray     = azArray[:n]
            elArray,n   = fUtilities.compress(elArray,good_for_fit,1)
            elArray     = elArray[:n]
            mapArray,n = fUtilities.compress(mapArray,good_for_fit,1)
            mapArray   = mapArray[:n]
            weightArray,n = fUtilities.compress(weightArray,good_for_fit,1)
            weightArray   = weightArray[:n]

        # now use the data to fit a 2D-Gaussian
        try:
            pointingResult = fitBaseEllipticalGaussian(mapArray,azArray,elArray,\
                                                       err = 1/sqrt(weightArray),\
                                                       fwhm=fwhm,\
                                                       gradient=gradient,\
                                                       circular=circular,\
                                                       Xpos=Xpos,Ypos=Ypos,\
                                                       fixedPos=fixedPos)
        except BoaError, error:
            self.__MessHand.warning('fit did not converge:')
            self.__MessHand.warning(error.msg)
            pointingResult = -1

        # TODO: is it the best thing to return the result?
        return pointingResult
    
            
    #--------------------------------------------------------------------------
    #
    # methods to compute rms/beam
    #--------------------------------------------------------------------------
    def rmsDistribution(self,cell=3):
        """
        DES: compute and plot the distribution of rms in the map
        INP: (int)  cell : size of cells on which rms are computed (default: 3x3)
        """
        
        data = self.Data
        WCS  = self.WCS
        
        allRms = []
        for x in range(WCS['NAXIS1']-cell):
            for y in range(WCS['NAXIS2']-cell):
                listVal = self.Data[x:x+cell,y:y+cell]
                listVal = ravel(listVal)
                    
                listVal, nNan = compressNan([listVal])
                if len(listVal) > 0:
                    mean = fStat.f_mean(listVal)
                    val = fStat.f_rms(listVal,mean)
                    allRms.append(val)

        return allRms
                    
    def meanDistribution(self,cell=3,limitsX=[],limitsY=[]):
        """
        DES: compute and plot the distribution of means in the map
        INP: (int)  cell : size of cells on which mean values are computed (default: 3x3)
             (i lists) limitsX/Y: optionally define a sub-region (pixel coord)
        """
        
        WCS  = self.WCS
        
        # Number of beams in the map
        nbX = WCS['NAXIS1']-cell
        nbY = WCS['NAXIS2']-cell

        # if sub-region asked, check that coord are within the map
        if limitsX:
            if limitsX[0] < 0:
                limitsX[0] = 0
            if limitsX[1] > nbX:
                limitsX[1] = nbX
        if limitsY:
            if limitsY[0] < 0:
                limitsY[0] = 0
            if limitsY[1] > nbY:
                limitsY[1] = nbY

        theData = copy.copy(self.Data)
        if limitsX:
            theData = theData[limitsX[0]:limitsX[1],:]
            nbX = limitsX[1]-limitsX[0]
        if limitsY:
            theData = theData[:,limitsY[0]:limitsY[1]]
            nbY = limitsY[1]-limitsY[0]

        allMean,nbOk = fStat.meandistribution(theData,nbX,nbY,nbX*nbY,cell)
        allMean = allMean[:nbOk]
        
        return allMean

    def computeRmsBeam(self,cell=3,rmsKappa=2,limitsX=[],limitsY=[]):
        """
        DES: compute rms/beam in a map (smoothed at beam resolution)
        INP: (f)     cell: size of one beam in pixel
             (f) rmsKappa: for kappa-sigma clipping before computing rms
             (i lists) limitsX/Y: optionally define a sub-region (pixel coord)
        """
        # Distribution of flux/beam
        m = self.meanDistribution(cell=cell,limitsX=limitsX,limitsY=limitsY)
        # Remove the NaNs at this stage
        mm, nNan = compressNan([m])
        # clip data at more than k-sigma (e.g. a source)
        m2,nbOk = fStat.clipping(mm,rmsKappa)
        m2 = m2[:nbOk]
        self.RmsBeam = fStat.f_stat(m2)[1]
        self.__MessHand.info("r.m.s. / beam = %f"%(self.RmsBeam))

    def computeRms(self,rmsKappa=2,limitsX=[],limitsY=[]):
        """
        DES: compute rms/beam in a map (dispersion between pixels)
        INP: (f) rmsKappa: for kappa-sigma clipping before computing rms
             (i lists) limitsX/Y: optionally define a sub-region (pixel coord)
        """
        if limitsX or limitsY:
            if not limitsX:
                localData = self.Data[:,limitsY[0]:limitsY[1]]
            elif not limitsY:
                localData = self.Data[limitsX[0]:limitsX[1],:]
            else:
                localData = self.Data[limitsX[0]:limitsX[1],
                                      limitsY[0]:limitsY[1]]
        else:
            localData = self.Data
        # Distribution of pixel values
        pixels = ravel(localData)
        # Remove the NaNs at this stage
        pixels,nb = compressNan([pixels])
        Mean,SDev,Med,MDev = fStat.f_stat(pixels)
        # clip data at more than k-sigma (e.g. a source)
        m2,nbOk = fStat.clipping(pixels,rmsKappa)
        m2 = m2[:nbOk]
        self.RmsBeam = fStat.f_stat(m2)[3]  # median deviation
        self.__MessHand.info("map r.m.s. = %f"%(self.RmsBeam))

    def writeFits(self,outfile='boaMap.fits', \
                  overwrite=0,limitsX=[],limitsY=[],intensityUnit="Jy/beam"):
        """
        DES: store the current map (2D array with WCS info) to a FITS file
        INP: (str)   outfile: output file name (default boaMap.fits)
             (bool) overwrite: overwrite existing file -
                              default = 0: do not overwrite existing file
             (f list) limitsX/Y: optional map limits (in world coordinates)
             (string) intensityUnit: optional unit of the intensity (default: "Jy/beam")
        """

        if os.path.exists(outfile):
            if not overwrite:
                self.__MessHand.error('File %s exists' % outfile)
                return
        
        try:
            dataset = BoaFits.createDataset("!" + outfile)
        except Exception, data:
            self.__MessHand.error('Could not open file %s: %s' % (outfile, data))
            return

        localMap = copy.deepcopy(self)
        if limitsX or limitsY:
            if not limitsX:
                # only limitsY given
                x1,x2 = 0,self.WCS['NAXIS1']
                i,y1 = self.wcs2pix(self.WCS['CRVAL1'],min(limitsY))
                i,y2 = self.wcs2pix(self.WCS['CRVAL1'],max(limitsY))
            elif not limitsY:
                # only limitsX given
                y1,y2 = 0,self.WCS['NAXIS2']
                x1,j = self.wcs2pix(max(limitsX),self.WCS['CRVAL2'])
                x2,j = self.wcs2pix(min(limitsX),self.WCS['CRVAL2'])
            else:
                # both limits given by user
                x1,y1 = self.wcs2pix(max(limitsX),min(limitsY))
                x2,y2 = self.wcs2pix(min(limitsX),max(limitsY))

            # make sure that the limits don't go outside the existing data
            x1 = max([0,int(x1)])
            x2 = min([self.WCS['NAXIS1'],int(x2)+1])
            y1 = max([0,int(y1)])
            y2 = min([self.WCS['NAXIS2'],int(y2)+1])
            localMap.Data     = self.Data[x1:x2,y1:y2]
            localMap.Weight   = self.Weight[x1:x2,y1:y2]
            localMap.Coverage = self.Coverage[x1:x2,y1:y2]
            # update WCS info - need to compute the actual limits
            limX1,limY1 = self.wcs2phy(x1,y1)
            limX2,limY2 = self.wcs2phy(x2,y2)
            minmaxXY = [limX2,limX1,limY1,limY2]
            pixSize = abs(self.WCS['CDELT1'])
            localMap.computeWCS(pixSize,minmax=minmaxXY)

        try:
            localMap.__writeImage(dataset, "Intensity", intensityUnit=intensityUnit)
            localMap.__writeImage(dataset, "Weight")
            localMap.__writeImage(dataset, "Coverage")
            dataset.close()
        except Exception, data:
            self.__MessHand.error('Could not write data to file %s: %s' % (outfile, data))
            return

        localMap = 0  # free memory
        
    def __writeImage(self, dataset, extname, intensityUnit=""):

        if extname == "Intensity":
            data = self.Data
        elif extname == "Weight":
            data = self.Weight
        elif extname == "Coverage":
            data = self.Coverage
        else:
            self.__MessHand.error("Don't know how to write image '%s'" % extname)

        if not data:
            self.__MessHand.info("No data to write image %s" % extname)
            return

        WCS = self.WCS
        
        try:
            dataImage = dataset.createImage(data)
            if not dataImage:
                self.__MessHand.error("Error while creating image '%s'" % extname)
                return

            dataImage.createKeywordDate()
            dataImage.createKeyword("CREATOR","BoA",
                                    comment="Bolometer Data Analysis Project")
            dataImage.createKeyword("EXTNAME",extname,
                                    comment="Type of data contained in this image")

            dataImage.createKeyword("MJD-OBS",WCS["MJD-OBS"],
                                    comment="time of observation")
            dataImage.createKeyword("RADESYS",WCS["RADESYS"],
                                    comment="frame of reference")
            dataImage.createKeyword("EQUINOX",WCS["EQUINOX"],
                                    comment="coordinate epoch")

            dataImage.createKeyword("CTYPE1",WCS["CTYPE1"],
                                    comment="Type of coordinate 1")
            dataImage.createKeyword("CRPIX1",WCS["CRPIX1"],
                                    comment="Reference pixel of coordinate 1")
            dataImage.createKeyword("CDELT1",WCS["CDELT1"],
                                    comment="Increment per pixel of coordinate 1")
            dataImage.createKeyword("CRVAL1",WCS["CRVAL1"],
                                    comment="Value of coordinate 1 at reference point")
            dataImage.createKeyword("CUNIT1",WCS["CUNIT1"],
                                    comment="Unit of coordinate 1")
            
            dataImage.createKeyword("CTYPE2",WCS["CTYPE2"],
                                    comment="Type of coordinate 2")
            dataImage.createKeyword("CRPIX2",WCS["CRPIX2"],
                                    comment="Reference pixel of coordinate 2")
            dataImage.createKeyword("CDELT2",WCS["CDELT2"],
                                    comment="Increment per pixel of coordinate 2")
            dataImage.createKeyword("CRVAL2",WCS["CRVAL2"],
                                    comment="Value of coordinate 2 at reference point")
            dataImage.createKeyword("CUNIT2",WCS["CUNIT2"],
                                    comment="Unit of coordinate 2")

            dataImage.createKeyword("PC1_1",WCS["PC1_1"],
                                    comment="Linear transformation matrix")
            dataImage.createKeyword("PC1_2",WCS["PC1_2"],
                                    comment="Linear transformation matrix")
            dataImage.createKeyword("PC2_1",WCS["PC2_1"],
                                    comment="Linear transformation matrix")
            dataImage.createKeyword("PC2_2",WCS["PC2_2"],
                                    comment="Linear transformation matrix")
            
            dataImage.createKeyword("OBJECT",self.Header["Object"],
                                    comment="Object observed")
            dataImage.createKeyword("TELESCOP",self.Header["Telescope"],
                                    comment="Telescope name")
            dataImage.createKeyword("FEBE",self.Header["FeBe"],
                                    comment="Frontend-backend combination")

            # BUNIT: this assumes that calibration was done!
            if extname == "Intensity":
                dataImage.createKeyword("BUNIT",intensityUnit,
                                        comment="Physical unit of image")
            dataImage.createKeyword("BMAJ",self.BeamSize,
                                    comment="Beam major axis")
            dataImage.createKeyword("BMIN",self.BeamSize,
                                    comment="Beam minor axis")
            dataImage.createKeyword("BPA",0.,
                                    comment="Beam position angle")

            dataImage.writeImage()
            
            self.__MessHand.longinfo("Image '%s' written. Size: %s"
                                   % (extname, str(dataImage.getShape())))
            
        except Exception, data:
            self.__MessHand.error("Error while writing image '%s'" % extname)
            self.__MessHand.error("Exception: '%s'" % data)
            return
        
    # -------------------------------------------------------------------
    def dumpMap(self,fileName='BoaMap.sav'):
        """
        DES: save an Image instance to a file
        INP: (string) fileName: name of the output file
                      (default = 'BoaMap.sav')
        """
        try:
            f = file(BoaConfig.outDir+fileName,'w')
        except:
            self.__MessHand.error(" permission denied, please change outdir")
            return
        cPickle.dump(self,f,2)
        f.close()
        self.__MessHand.info(" current Image successfully written to %s"%fileName)


#----------------------------------------------------------------------------------
#----- Kernel Class ---------------------------------------------------------------
#----------------------------------------------------------------------------------
class Kernel(Image):
   """
   NAM: Kernel (class)
   DES: define a kernel
   """

   def __init__(self, pixelSize, beamSize):
	"""
	DES: Initialise an instance of a Kernel class
	INP: (float) pixelSize: the physical size of a pixel
             (float) beamSize : the beam FWHM in the same unit
	"""
	
	Image.__init__(self)
	
	self.computeWCS(pixelSize,\
			minmax = array([-1,1,-1,1],Float32)*beamSize*3.)
	kernel_azimuth, kernel_elevation = self.physicalCoordinates()

	#kernel_parameters = array([0.,0.,0.,beamSize**2*16*pi*log(2.),
	kernel_parameters = array([0.,0.,0.,1.,
				   0.,0.,beamSize,beamSize,0.])
	tmpData = modelBaseEllipticalGaussian(kernel_parameters,\
                                              [kernel_azimuth,kernel_elevation])
        self.Data = tmpData.astype(Float32)

	
#----------------------------------------------------------------------------------
#----- Boa Mapping Class ----------------------------------------------------------
#----------------------------------------------------------------------------------
class Map(BoaDataAnalyser.DataAna):
    """
    NAM: Map (class)
    DES: An object of this class is responsible for the restoration of
         mapping data of single or multiple files.
    """
    
    def __init__(self):
        """
        DES: Initialise an instance.
        """
        BoaDataAnalyser.DataAna.__init__(self)
        
        self.__Results = []
	self.Map = Image()
        

#--------------------------------------------------------------------------------
#----- map methods --------------------------------------------------------------
#--------------------------------------------------------------------------------

    def showMap(self,style='idl4',wedge=1,\
                limitsZ=[],aspect=1,limitsX=[],limitsY=[],caption=None,\
                doContour=0,levels=[],showRms=1,rmsKappa=2):
        """
        DES: show the reconstructed map in (Az,El) or (Ra,Dec)
        """

        if self.Map.Data == []:
            self.MessHand.error('No map computed yet')
            return
	if not caption:
            caption = self.ScanParam.caption()
            
	self.Map.display(style=style,
			 limitsX=limitsX,limitsY=limitsY,limitsZ=limitsZ,
			 aspect=aspect,caption=caption,wedge=wedge,
			 doContour=doContour,levels=levels,
                         showRms=showRms, rmsKappa=rmsKappa)

    #--------------------------------------------------------------------------------

    def displayMap(self, weight=0, coverage=0, style='idl4', caption='', wedge=1, \
		aspect=1, overplot=0, doContour=0, levels=[], labelContour=0,\
		limitsX = [], limitsY = [], limitsZ = [], \
                showRms=0, rmsKappa=2 ):

        if self.Map.Data == []:
            self.MessHand.error('No map computed yet')
            return

        self.Map.display(weight=weight, coverage=coverage, style=style, \
                         caption=caption, wedge=wedge, aspect=aspect, \
                         overplot=overplot, doContour=doContour, levels=levels, \
                         labelContour=labelContour, \
                         limitsX = limitsX, limitsY = limitsY, limitsZ = limitsZ, \
                         showRms=showRms, rmsKappa=rmsKappa )

    #--------------------------------------------------------------------------------

    def smoothMap(self,Size):
       """
       DES: smooth the image with a 2D gaussian of given FWHM
       INP: (float) Size : the FWHM of the smoothing gaussian
       """
       
       if self.Map.Data == []:
           self.MessHand.error('No map computed yet')
           return

       self.Map.smoothBy(Size)

    #--------------------------------------------------------------------------------

    def getPixelFromMap(self,nbPix=3):
        """
        DES: allow user to get pixel values using mouse
        INP: (int)  nbPix : size of area to compute average (default 3x3)
        """
       
        if self.Map.Data == []:
            self.MessHand.error('No map computed yet')
            return

        self.Map.getPixel(nbPix)

    #--------------------------------------------------------------------------------

    def computeRmsFromMap(self,rmsKappa=2,limitsX=[],limitsY=[]):
        """
        DES: compute rms/beam in a map (dispersion between pixels)
        INP: (f) rmsKappa: for kappa-sigma clipping before computing rms
             (i lists) limitsX/Y: optionally define a sub-region (pixel coord)
        """
       
        if self.Map.Data == []:
            self.MessHand.error('No map computed yet')
            return

        self.Map.computeRms(rmsKappa,limitsX,limitsY)

#--------------------------------------------------------------------------------
#----- public methods -----------------------------------------------------------
#--------------------------------------------------------------------------------

    def doMap(self,chanList=[], \
              channelFlag=[], plotFlaggedChannels=0, \
              dataFlag=[], plotFlaggedData=0, \
              oversamp=2.0,beammap = 0, system='HO',\
              sizeX=[],sizeY=[],limitsZ=[],style='idl4',wedge=1,smooth=0,noPlot=0,\
              caption=None,aspect=1,showRms=1,rmsKappa=2.,derotate=0):

        """
        DES: reconstruct a map in (Az,El) coordinates combining bolometers
        INP: (int list) chanList = channels to consider
             (integer list) channelFlag : plot data from channels flagged or unflagged accordingly
             (log)  plotFlaggedChannels : channelFlag revers to flagged/unflagged data
             (integer list)    dataFlag : plot data flagged or unflagged accordingly
             (log)      plotFlaggedData : dataFlag revers to flagged/unflagged data
                                          flag   | plotFlagged | Plot..
                                          'None' |  0          | all data
                                          []     |  0          | unflagged data (default)
                                          []     |  1          | data with at least one flag set
                                          1      |  0          | data with flag 1 not set
                                          1      |  1          | data with flag 1 set
                                          [1,2]  |  0          | data with neither flag 1 nor flag 2 set
                                          [1,2]  |  1          | data with either flag 1 or flag 2 set
             (float)    oversamp = oversampling factor (beam fwhm / pixel size). Default=2.
             (log)       beammap = compute a beam map (default: no)
             (str)        system = coord. system, one of 'HO' (Az,El *offsets*) or 'EQ'
                                   (RA, Dec absolute coord.) - default = 'HO'
                                   optionally 'EQFAST' to do only one rotation
                                   on small maps (faster)
             (list float)  sizeX = limits in Az of the map
             (list float)  sizeY = limits in El of the map
             (logical)     noNan = remove NaN in self.Results?
             (str)         style = color table to use in image
             (logical)    smooth = do we smooth with beam? (default: no)
	     (logical)    noPlot = do not plot the map? (default: no, i.e. yes we plot)
	     (str)       caption = plot caption
             (logical)    aspect = keep aspect ratio? (default: yes)
             (logical)   showRms = compute and print rms/beam? (default: yes)
             (float)    rmsKappa = kappa in kappa-sigma clipping used to compute rms
             (int)      derotate = derotate Nasmyth array by Elevation
        """
        # check channel list; will return all non-flagged channels if chanList = []
        chanList = self.BolometerArray.checkChanList(chanList, \
                                                     flag=channelFlag,getFlagged=plotFlaggedChannels)

        if not(len(chanList)):
            self.MessHand.error("no valid channels")
            return
        chanListIndexes = self.BolometerArray.getChanIndex(chanList)

        myTiming = Timing()
        myTiming.setTime()

        # Select offsets according to system, 'HO' or 'EQ'
        system = string.upper(system)
        if system == 'HO':
            if derotate:
                ElAverage = average(self.ScanParam.El)
                self.BolometerArray.rotateArray(derotate*ElAverage)
            XYOffsets = array([self.ScanParam.get('AzimuthOffset',flag='None'), \
                               self.ScanParam.get('ElevationOffset',flag='None')])
            XOffsets = self.ScanParam.get('AzimuthOffset')
            YOffsets = self.ScanParam.get('ElevationOffset')
        elif (system == 'EQ') or (system=='EQFAST'):
            XYOffsets = array([self.ScanParam.get('RA',flag='None'), \
                               self.ScanParam.get('Dec',flag='None')])
            XOffsets = self.ScanParam.get('RA')
            YOffsets = self.ScanParam.get('Dec')
            
 	data        = self.Data
	dataWeights = self.DataWeights
       
	OffsetsUsed = array(self.BolometerArray.getChanSep(self.BolometerArray.UsedChannels))
        
        if beammap:
            OffsetsUsed = OffsetsUsed * 0.0
            if system == 'EQ':
                self.MessHand.warning('Doing a beam map in EQ system probably does not make sense')

        # Retrieve the beam and compute pixel size
        fwhm = self.BolometerArray.BeamSize
        pixelSize = fwhm / oversamp

        if (system == 'EQ') or (system == 'EQFAST'):
            pixelSize = pixelSize/3600.
            fwhm = fwhm/3600.
            
            if (system == 'EQ'):
                if (derotate):
                    rotAngles = array(self.ScanParam.ParAngle)+derotate*array(self.ScanParam.El)
                else:
                    rotAngles = array(self.ScanParam.ParAngle)
                chanListAzEl = array(self.BolometerArray.UsedChannels)-1
                OffsetsAzEl=array((take(self.BolometerArray.Offsets[0,:],chanListAzEl), \
                                   take(self.BolometerArray.Offsets[1,:],chanListAzEl)))

                refChOffsets=self.BolometerArray.Offsets[:,self.BolometerArray.RefChannel-1]

            # even if system is 'EQ', the following needs to be done in order
            # to allow computation of the extrema of the map
            OffsetsUsed = OffsetsUsed / 3600.
            Yrange = fStat.minmax(YOffsets)
            OffsetsUsed[0,:] /= array(cos((Yrange[0]+Yrange[1])/2. * pi / 180.),'f')
            
        # compute roughly the extrema of the map (taking into account the ScanParam
        # flags but not the BolometerArray Flags nor the Data Flags
	minmaxXY = concatenate(\
		(fStat.minmax(XOffsets)+array([-1,1])*2*pixelSize + \
		 fStat.minmax(take(OffsetsUsed[0,:],chanListIndexes)), \
		 fStat.minmax(YOffsets)+array([-1,1])*2*pixelSize + \
		 fStat.minmax(take(OffsetsUsed[1,:],chanListIndexes)) ) \
		)
        if (system == 'EQ') or (system == 'EQFAST'):
            # swap X min and max (East to the left)
            minmaxXY[0],minmaxXY[1] = minmaxXY[1],minmaxXY[0]
            
	# Create the resulting image
        Map = self.Map
        Map.__init__()

        Map.Header['Object']    = self.ScanParam.Object
        Map.Header['Telescope'] = self.BolometerArray.Telescope.Name
        Map.Header['FeBe']      = self.BolometerArray.FeBe
        Map.BeamSize = fwhm
        
        if system == 'HO':
            Map.WCS['CTYPE1']   = 'OLON--SFL'
            Map.WCS['CTYPE2']   = 'OLAT--SFL'
            Map.WCS['CUNIT1']   = 'arcsec'
            Map.WCS['CUNIT2']   = 'arcsec'
            cdeltUnit           = 1./3600.
        elif (system == 'EQ') or (system == 'EQFAST'):
            Map.WCS['CTYPE1']   = 'RA---SFL'
            Map.WCS['CTYPE2']   = 'DEC--SFL'
            Map.WCS['CUNIT1']   = 'deg'
            Map.WCS['CUNIT2']   = 'deg'
            cdeltUnit           = 1.
        Map.WCS['MJD-OBS']  = self.ScanParam.DateObs
        
        # update main WCS keywords
        Map.computeWCS(pixelSize,sizeX,sizeY,minmaxXY)
	AXIS1 = array([Map.WCS['NAXIS1'],Map.WCS['CRPIX1'],Map.WCS['CDELT1'],
                       Map.WCS['CRVAL1'],cdeltUnit])
	AXIS2 = array([Map.WCS['NAXIS2'],Map.WCS['CRPIX2'],Map.WCS['CDELT2'],
                       Map.WCS['CRVAL2'],cdeltUnit])

        self.MessHand.longinfo("Building a map with dimensions (x,y) = "+\
                           str(int(AXIS1[0]))+','+str(int(AXIS2[0])))

        if Map.WCS['NAXIS1']*Map.WCS['NAXIS2']*32*3/8/1e6 > 600:
            if BoaConfig.online:
                self.MessHand.error('Such maps would require more than 600Mb of memory, aborting...')
                return
            elif not self.MessHand.yesno('do you really want to do such a big map (>600Mb) ?'):
                return
            
        mapData     = zeros((int(AXIS1[0]),int(AXIS2[0])),typecode=Float32)
        mapWeight   = zeros((int(AXIS1[0]),int(AXIS2[0])),typecode=Float32)
        mapCoverage = zeros((int(AXIS1[0]),int(AXIS2[0])),typecode=Float32)

        dataFlag = self._removeReservedFlagValues(dataFlag, self.rflags['CHANNEL FLAGGED'])
        if dataFlag==None:
            self.MessHand.error("no valid flags")
            return

        if dataFlag in ['','None']:
            dataFlagMask = ones(shape=self.DataFlags.shape, typecode=self.DataFlags.typecode())
        else:
            if plotFlaggedData:
                dataFlagMask = self._flagHandler.isSetMask(self.DataFlags, dataFlag)
            else:
                dataFlagMask = self._flagHandler.isUnsetMask(self.DataFlags, dataFlag)

        if (system=='EQFAST') or ((system=='HO')):
            mapData, mapWeight, mapCoverage = fMap.horizontalprojection(chanListIndexes, data, \
                                                                        dataWeights, \
                                                                        dataFlagMask, 1, \
                                                                        XYOffsets, OffsetsUsed,
                                                                        AXIS1, AXIS2,
                                                                        mapData, mapWeight, mapCoverage)
        if system=='EQ':
            mapData, mapWeight, mapCoverage = fMap.equatorialprojection(chanListIndexes, data, \
                                                                        dataWeights, \
                                                                        dataFlagMask, 1, \
                                                                        XYOffsets, OffsetsAzEl,
                                                                        rotAngles,
                                                                        refChOffsets,
                                                                        AXIS1, AXIS2,
                                                                        mapData, mapWeight, mapCoverage) 
       	Map.Data     = mapData
        Map.Weight   = mapWeight
        Map.Coverage = mapCoverage
        
	self.Map = Map

	# Smoothing after normalisation 
        if smooth:
	    kernel_to_smooth = Kernel(pixelSize,fwhm)
	    self.Map.smoothWith(kernel_to_smooth)

	if not noPlot:
            self.showMap(wedge=wedge,style=style,aspect=aspect,
                         limitsZ=limitsZ,caption=caption,
                         showRms=showRms,rmsKappa=rmsKappa)

        self.MessHand.debug(" map done in "+ str(myTiming))

        if ((system=='HO') and derotate):
            self.BolometerArray.rotateArray(-derotate*ElAverage)   # rotate array back


    #--------------------------------------------------------------------------------
    def zoom(self,mouse=1,style='idl4',wedge=1,\
                limitsZ=[],aspect=1,limitsX=[],limitsY=[],caption=None,\
                doContour=0,levels=[],showRms=1,rmsKappa=2):
        """
        DES: allow the user to select a region in the map to zoom in
        INP: (bool) mouse: use the mouse? (default: yes)
               (other parameters: same as showMap)
        """

	if not caption:
            caption = self.ScanParam.caption()
            
        self.Map.zoom(mouse=mouse,style=style,
                      limitsX=limitsX,limitsY=limitsY,limitsZ=limitsZ,
                      aspect=aspect,caption=caption,wedge=wedge,
                      doContour=doContour,levels=levels,
                      showRms=showRms, rmsKappa=rmsKappa)
        
    #--------------------------------------------------------------------------------
    def beamMap(self,chanList=[], \
                channelFlag=[], plotFlaggedChannels=0, \
                dataFlag=[], plotFlaggedData=0, \
                oversamp=2.0,sizeX=[],sizeY=[],\
		style='idl4'):
        """
        DES: build a beam map in (Az,El) coordinates
        INP: (int list) chanList = channels to consider
             (integer list) channelFlag : plot data from channels flagged or unflagged accordingly
             (log)  plotFlaggedChannels : channelFlag revers to flagged/unflagged data
             (integer list)    dataFlag : plot data flagged or unflagged accordingly
             (log)      plotFlaggedData : dataFlag revers to flagged/unflagged data
                                          flag   | plotFlagged | Plot..
                                          'None' |  0          | all data
                                          []     |  0          | unflagged data (default)
                                          []     |  1          | data with at least one flag set
                                          1      |  0          | data with flag 1 not set
                                          1      |  1          | data with flag 1 set
                                          [1,2]  |  0          | data with neither flag 1 nor flag 2 set
                                          [1,2]  |  1          | data with either flag 1 or flag 2 set
             (float) oversamp = oversampling factor (beam fwhm / pixel size). Default=2.
             (list float) sizeX = limits in Az of the map
             (list float) sizeY = limits in El of the map
        """

        # do exactly as a normal map, but don't add channel offsets
        self.doMap(chanList=chanList, \
                   channelFlag=channelFlag, plotFlaggedChannels=plotFlaggedChannels, \
                   dataFlag=dataFlag, plotFlaggedData=plotFlaggedData, \
                   oversamp=oversamp,\
		   beammap=1,sizeX=sizeX,sizeY=sizeY,style=style)


    #----------------------------------------------------------------------------
    def chanMap(self,chanList=[], \
                channelFlag=[], plotFlaggedChannels=0, \
                dataFlag=[], plotFlaggedData=0, \
                oversamp=1.,sizeX=[],sizeY=[],\
                style='idl4',limitsZ=[],center=0,showRms=0,rmsKappa=2.):
        """
        DES: Compute and plot channel maps in HO offset coordinates
        INP: (i list) chanList = channels to consider
             (integer list) channelFlag : plot data from channels flagged or unflagged accordingly
             (log)  plotFlaggedChannels : channelFlag revers to flagged/unflagged data
             (integer list)    dataFlag : plot data flagged or unflagged accordingly
             (log)      plotFlaggedData : dataFlag revers to flagged/unflagged data
                                          flag   | plotFlagged | Plot..
                                          'None' |  0          | all data
                                          []     |  0          | unflagged data (default)
                                          []     |  1          | data with at least one flag set
                                          1      |  0          | data with flag 1 not set
                                          1      |  1          | data with flag 1 set
                                          [1,2]  |  0          | data with neither flag 1 nor flag 2 set
                                          [1,2]  |  1          | data with either flag 1 or flag 2 set
             (float)  oversamp = oversampling factor (beam fwhm / pixel size). Default=2.
             (2xfloat)   sizeX = limits in Az of the map
             (2xfloat)   sizeY = limits in El of the map
             (str)       style = color table to use in images
             (logical)  center = if set, it will shift each map by the bolometer offsets.
	                Thereby it shifts the source to the center of each channel map.
             (logical) showRms = compute and print rms/beam? (default: no)
             (float)  rmsKappa = kappa in kappa-sigma clipping used to compute rms
        """

        # Check the list of input channels
        chanList = self.BolometerArray.checkChanList(chanList, \
                                                     flag=channelFlag,getFlagged=plotFlaggedChannels)

        if not(len(chanList)):
            self.MessHand.error("no valid channel")
            return

	chanListIndexes = self.BolometerArray.getChanIndex(chanList)

	data        = self.Data
	dataWeights = self.DataWeights
	AzElOffsets = array([self.ScanParam.get('AzimuthOffset',flag='None'), \
			     self.ScanParam.get('ElevationOffset',flag='None')])

     
	OffsetsUsed = array(self.BolometerArray.getChanSep(self.BolometerArray.UsedChannels))
	#OffsetsUsed     = array([zeros(self.BolometerArray.NUsedChannels,Float32),\
        #                         zeros(self.BolometerArray.NUsedChannels,Float32)],Float32)

        if center:
            OffsetsUsed = OffsetsUsed*0

        # Retrieve the beam
        fwhm = self.BolometerArray.BeamSize

        # Compute the pixel size and ...
        pixelSize = fwhm / oversamp
        
        # compute roughly the extrema of the map (taking into account
        # the ScanParam flags but not the BolometerArray Flags or the
        # Data Flags
	
	AzOffsets = self.ScanParam.get('AzimuthOffset')
	ElOffsets = self.ScanParam.get('ElevationOffset')

	minmaxAzEl = concatenate(\
		(fStat.minmax(AzOffsets)+array([-1,1])*2*pixelSize - \
		 fStat.minmax(take(OffsetsUsed[0,:],chanListIndexes))[::-1], \
		 fStat.minmax(ElOffsets)+array([-1,1])*2*pixelSize - \
		 fStat.minmax(take(OffsetsUsed[1,:],chanListIndexes))[::-1] ) \
		)

	# Create a master image
        Map = Image()

        Map.Header['Object']    = self.ScanParam.Object
        Map.Header['Telescope'] = self.BolometerArray.Telescope.Name
        Map.Header['FeBe']      = self.BolometerArray.FeBe

        Map.WCS['CTYPE1']   = 'OLON--SFL'
        Map.WCS['CTYPE2']   = 'OLAT--SFL'
        Map.WCS['CUNIT1']   = 'arcsec'
        Map.WCS['CUNIT2']   = 'arcsec'
        Map.WCS['MJD-OBS']  = self.ScanParam.DateObs
        
        # update main WCS keywords
        Map.computeWCS(pixelSize,sizeX,sizeY,minmaxAzEl)
        cdeltUnit = 1./3600.  # CDELTi are in arcsec
	AXIS1 = array([Map.WCS['NAXIS1'],Map.WCS['CRPIX1'],Map.WCS['CDELT1'],
                       Map.WCS['CRVAL1'],cdeltUnit])
	AXIS2 = array([Map.WCS['NAXIS2'],Map.WCS['CRPIX2'],Map.WCS['CDELT2'],
                       Map.WCS['CRVAL2'],cdeltUnit])

        self.MessHand.info("Building maps with dimensions (x,y) = "+str(int(AXIS1[0]))+','+str(int(AXIS2[0])))

        # Initialise a list of maps
        allMaps = []
        
	ChanRef = self.BolometerArray.RefChannel
      
        # Now start a loop on channels, to build each map
        for chan in range(len(chanList)):
            mapData     = zeros((int(AXIS1[0]),int(AXIS2[0])),typecode=Float32)
            mapWeight   = zeros((int(AXIS1[0]),int(AXIS2[0])),typecode=Float32)
            mapCoverage = zeros((int(AXIS1[0]),int(AXIS2[0])),typecode=Float32)
            
            dataFlag = self._removeReservedFlagValues(dataFlag, self.rflags['CHANNEL FLAGGED'])
            if dataFlag==None:
                self.MessHand.error("no valid flags")
                return

            if dataFlag in ['','None']:
                dataFlagMask = ones(shape=self.DataFlags.shape, typecode=self.DataFlags.typecode())
            else:
                if plotFlaggedData:
                    dataFlagMask = self._flagHandler.isSetMask(self.DataFlags, dataFlag)
                else:
                    dataFlagMask = self._flagHandler.isUnsetMask(self.DataFlags, dataFlag)

            mapData, mapWeight, mapCoverage = fMap.horizontalprojection(chanListIndexes[chan], data, \
                                                                        dataWeights, \
                                                                        dataFlagMask, 1, \
                                                                        AzElOffsets, OffsetsUsed,
                                                                        AXIS1, AXIS2,
                                                                        mapData, mapWeight, mapCoverage)
            # Storing of the map
            allMaps.append(copy.copy(mapData))
            
        # Store all maps in Results attribute
        self.__Results = allMaps
    
        labelX="\gD Az ['']"
        labelY="\gD El ['']"
        
        MultiPlot.draw(chanList,allMaps,WCS=Map.WCS,\
		       wedge=1,labelX=labelX, labelY=labelY,\
		       caption=self.ScanParam.caption(),style=style,\
		       limitsZ=limitsZ,nan=1)
        if showRms:
            cell = oversamp
            nbX = int(AXIS1[0]) - cell
            nbY = int(AXIS2[0]) - cell
            for chan in range(len(chanList)):
                oneMap = allMaps[chan]
                # Distribution of flux/beam
                m,nbOk = fStat.meandistribution(oneMap,nbX,nbY,nbX*nbY,cell)
                m = m[:nbOk]
                # clip data at more than 2-sigma (e.g. a source)
                m2,nbOk = fStat.clipping(m,rmsKappa)
                m2 = m2[:nbOk]
                oneRms = fStat.f_stat(m2)[1]
                self.MessHand.info("Channel %i : r.m.s. / beam = %f"%(chanList[chan],oneRms))

    #----------------------------------------------------------------------------
    def plotBoloRms(self,smoothFactor=1.5,style='idl4',limitsX=[],limitsY=[],limitsZ=[],
                    caption=''):
        """
        DES: plot the array with color scale showing rms
        INP: (float) smoothFactor: the map is smooted by this factor x beam
             style, limits? : plot parameters
        """
        chanList = self.BolometerArray.checkChanList([])
        if len(chanList)<1: 
            self.MessHand.error("no valid channel")
            return

        if not self._DataAna__statisticsDone:
            self._DataAna__statistics()

        bolo = self.BolometerArray
        rms = self.getChanListData('rms',chanList)

        # Create a Map object to produce the plot
        mapRms = Map()
        mapRms.BolometerArray = bolo
        nbInt  = 13
        nbBolo = bolo.NChannels
        mapRms.ScanParam.NInt = nbInt
        mapRms.ScanParam.NObs = 1
        b = bolo.BeamSize * 2. / 3600.
        mapRms.ScanParam.AzOff = array([0.,0.,0.   ,0.  ,0.,-b,-b/2.,b/2.,b ,b/2.,-b/2.,-b/2.,b/2.],'f')
        mapRms.ScanParam.ElOff = array([0.,-b,-b/2.,b/2.,b ,0.,0.   ,0.  ,0.,b/2.,b/2.,-b/2.,-b/2.],'f')
        mapRms.ScanParam.Flags = zeros((nbInt),UInt32)
        mapRms.Data        = ones((nbInt,nbBolo),'f')
        mapRms.DataWeights = ones((nbInt,nbBolo),'f')
        mapRms.DataFlags   = zeros((nbInt,nbBolo),UInt8)
        for i in range(len(chanList)):
            mapRms.Data[:,chanList[i]-1] *= array(rms[i],'f')

        mapRms.doMap(oversamp=1,noPlot=1,showRms=0)
        mapRms.Map.smoothBy(smoothFactor*bolo.BeamSize)
        # if no limitsZ input, try to compute a good one
        if limitsZ == []:
            med = fStat.f_median(rms)
            sigma = fStat.f_rms(rms,med)
            z1 = max(min(rms),med-sigma)
            z2 = min(max(rms),med+sigma)
            limitsZ = [z1,z2]
        if not caption:
            caption = bolo.FeBe + ' - channel rms'
        mapRms.Map.display(style=style,limitsX=limitsX,limitsY=limitsY,
                           limitsZ=limitsZ,caption=caption)
        bolo.plotArray(overplot=1,num=1)
        
    #--------------------------------------------------------------------------------
    def reduce(self,datasetName='',obstoProc=[],update=0):
        """
        DES: Process a map scan - this method is called by the apexCalibrator
        INP: (string) datasetName: path to the dataset to be reduced
             (i list) obstoProc: list of subscans to consider (default: all)
        """
        self.read(inFile=datasetName,subscans=obstoProc,update=update)

        # Automatic flagging of dead channels
        self.flagFractionRms(ratio=5.)   # flag at median(rms)/5 and median*5

        # Remove correlated noise - assume that ref. channel is good!
        # unflag it if it was flagged
        full = tolist_boa(self.BolometerArray.checkChanList([]))
        if not full.count(self.BolometerArray.RefChannel):
            self.unflagChannels(self.BolometerArray.RefChannel)
        self.medianNoiseRemoval()
        # Another zero-order baseline
        self.medianBaseline(subscan = 1)
        # compute weight - hopefully, no too strong source!
        self.computeWeight()
        self.doMap(oversamp=3)
        
#--------------------------------------------------------------------------------
def mapSum(mapList):
    """
    Function (NOT a method) to co-add Image objects.
    Map data, weights and coverage planes are co-added.
    Returns a new Image object, with same WCS and data size.
    
    WARNING: this function assumes that all Image objects correspond
             to the same region of the sky (same size, same center)

    # Example of use:
    scans   = [some list of scan numbers]
    mapList = []  # initialise empty list
    ra1,ra2,de1,de2 = ...  # define limits to be used for all maps
    for s in scans:
        read(str(s))
        <processing of each scan>
        mapping(system='EQ',sizeX=[ra1,ra2],sizeY=[de1,de2])
        mapList.append(data.Map)
    ms = mapSum(mapList)  # co-added Image object
    ms.display()          # can be displayed
    ms.zoom()             # zoom function can be used
    ms.writeFits("output.fits")
    """
    
    result = copy.deepcopy(mapList[0])
    nX = result.WCS['NAXIS1']
    nY = result.WCS['NAXIS2']
    for i in range(nX):
        for j in range(nY):
            Pij = 0.
            Wij = 0.
            Cij = 0.
            for m in mapList:
                if str(m.Data[i,j]) != str(float('nan')):
                    Pij += m.Weight[i,j] * m.Data[i,j]
                    Wij += m.Weight[i,j]
                    Cij += m.Coverage[i,j]
            if Wij:
                result.Data[i,j] = Pij/Wij
                result.Weight[i,j] = Wij
                result.Coverage[i,j] = Cij
    return result

