# Copyright (C) 2002-2006
# Max-Planck-Institut fuer Radioastronomie Bonn
# Argelander Institut fuer Astronomie
# Astronomisches Institut der Ruhr-Universitaet Bochum
#
# Produced for the LABOCA project
#
# This library is free software; you can redistribute it and/or modify it under
# the terms of the GNU Library General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) any
# later version.
#
# This library is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY FITNESS
# FOR A PARTICULAR PURPOSE. See the GNU Library General Public License for more
# details.
#
# You should have received a copy of the GNU Library General Public License
# along with this library; if not, write to the Free Software Foundation, Inc.,
# 675 Massachusetts Ave, Cambridge, MA 02139, USA.
#

"""
NAM: BoaPoint.py (module)
DES: contains the BoA Pointing reduction tools
"""
__version__=  '$Revision: 2355 $'
__date__=     '$Date: 2007-08-21 01:50:31 +0200 (Di, 21 Aug 2007) $'

#----------------------------------------------------------------------------------
#----- Import ---------------------------------------------------------------------
#----------------------------------------------------------------------------------
import BoaMapping
from Numeric import *

from boa           import BoaConfig
from boa.Bogli     import Plot, Forms
from boa.Utilities import array2list, fitBaseEllipticalGaussian, tolist_boa
from boa.BoaError  import BoaError
from boa.fortran   import fUtilities, fMap, fStat

#----------------------------------------------------------------------------------
#----- Boa Pointing Class ----------------------------------------------------------
#----------------------------------------------------------------------------------
class Point(BoaMapping.Map):
    """
    NAM: Point (class)
    DES: An object of this class is responsible for the reduction of
         pointing scan(s)
    """

  
    def __init__(self):
        """
        DES: Initialise an instance.
        """
        BoaMapping.Map.__init__(self)

        self.Maps = []
        self.__Result = []
        self.PointingResult = 0

#--------------------------------------------------------------------------------
#----- public methods -----------------------------------------------------------
#--------------------------------------------------------------------------------

    def iterMap(self,chanList=[],phase=0,flag=0,sizeX=[],sizeY=[]):
        """
        DES: reconstruct a map in (Az,El) coordinates combining bolometers
             and using varying scale to zoom on signal
        INP: (int list) chanList = channels to consider
             (int) phase = phase to plot
             (int) flag  = flag values to consider
             (list float) sizeX = limits in Az of the map
             (list float) sizeY = limits in El of the map
        """

        self.Maps = []  # reinitialise Maps attribute
        # start with one pixel per beam
	self.slowMap(chanList=chanList,phase=phase,flag=flag,sizeX=sizeX,\
		sizeY=sizeY,oversamp=1.)
       
        firstMap = self._Map__Results
	self.Maps.append(firstMap)

	# localize max. of signal
	dimX = shape(firstMap)[0]
	dimY = shape(firstMap)[1]

	# convert to list, to use index
	l = tolist_boa(ravel(firstMap))
        
        # remove NaN values, if any
        allNumbers = []
        for x in l:
              if str(x) != str(float('Nan')):
                    allNumbers.append(x)
                    
        maxi = max(allNumbers)
        mini = min(allNumbers)
        # remove the Nan in this list, otherwise index does not work
        for n in range(len(l)):
            if str(l[n]) == str(float('Nan')):
                   l[n] = mini
	posMax = l.index(maxi)
        
	# convert to x,y
	maxX = int(posMax/dimY)
	maxY = posMax - dimY * maxX + 0.5
        maxX = maxX + 0.5
 	# then to Az, El
        maxAz,maxEl = self.wcs2phy(maxX,maxY)
        # new limits for the second map
        newX1,newY1 = self.wcs2phy(maxX-3.5,maxY-3.5)
        newX2,newY2 = self.wcs2phy(maxX+3.5,maxY+3.5)
        Plot.plot([newX1,newX2,newX2,newX1,newX1],\
                  [newY1,newY1,newY2,newY2,newY1],style='l',overplot=1)
	self.MessHand.info(" after 1st iteration, offsets Az,El ="+str(maxAz)+","+str(maxEl))

	# second iteration:
	# compute a second map +/- 3 beams around max
        self.slowMap(chanList=chanList,phase=phase,flag=flag,\
		sizeX=[newX1,newX2],sizeY=[newY1,newY2],oversamp=3.)
        secondMap = self._Map__Results
	self.Maps.append(secondMap)
	# determine new maximum position
	dimX = shape(secondMap)[0]
	dimY = shape(secondMap)[1]
	# convert to list, to use index
	l = tolist_boa(ravel(secondMap))
        # remove NaN values, if any
        allNumbers = []
        for x in l:
              if str(x) != str(float('Nan')):
                    allNumbers.append(x)
        maxi = max(allNumbers)
        mini = min(allNumbers)
        # remove the Nan in this list, otherwise index does not work
        for n in range(len(l)):
            if str(l[n]) == str(float('Nan')):
                   l[n] = mini
	posMax = l.index(maxi)
	# convert to x,y
	maxX = int(posMax/dimY)
	maxY = posMax - dimY * maxX + 0.5
        maxX = maxX + 0.5
	# then to Az, El
        maxAz,maxEl = self.wcs2phy(maxX,maxY)
        # new limits for the third map
        newX1,newY1 = self.wcs2phy(maxX-3.5,maxY-3.5)
        newX2,newY2 = self.wcs2phy(maxX+3.5,maxY+3.5)
        Plot.plot([newX1,newX2,newX2,newX1,newX1],\
                  [newY1,newY1,newY2,newY2,newY1],style='l',overplot=1)
	self.MessHand.info(" after 2nd iteration, offsets Az,El ="+str(maxAz)+","+str(maxEl))
		
	# third iteration:
	# compute a second map +/- 3 pixels around max
        self.slowMap(chanList=chanList,phase=phase,flag=flag,\
		sizeX=[newX1,newX2],sizeY=[newY1,newY2],oversamp=5.)

	self.Maps.append(self._Map__Results)
		
    def solvePointing(self,chanList=[], gradient=0, circular=0, radius= -5, \
                      Xpos = 0., Ypos = 0., fixedPos = 0, \
                      plot=0, display=1, caption='', aspect=1):
        """
        DES: compute the offset
        INP: (int list) chanList: list of channels to be used (default: all)
             (boolean)  gradient: shall we fit a gradient ? (default: no)
             (boolean)  circular: fit a cricular gaussian instead of an elliptical gaussian
             (float)    radius  : use only bolo inside this radius (negative means multiple of beam) (default 5 beams)
             (float)  Xpos,Ypos : source position if using fixed position
             (boolean) fixedPos : if set, don't fit position, but use Xpos, Ypos
             (boolean)  plot    : do we plot the results? (default: no)
             (boolean)  display    : display the result of the fit (default: yes)    
        OUT: store in self.PoitingResult the results of the fit (i.e. all parameters
             as computed by mpfit routine). If mpfit failed, then self.PoitingResult
             is set to -1
        """

        # Retrieve the data..
        # The az/el off already contains the actual array offsets 
        chanList = self.BolometerArray.checkChanList(chanList)
        theData = self.getChanListData('flux',chanList)
        dataX = self.getChanListData('azoff',chanList)
        dataY = self.getChanListData('eloff',chanList)

        fwhm = self.BolometerArray.BeamSize

        # Transform that to a 1D array
        dataX = array2list(dataX)
        dataY = array2list(dataY)
        theData = array2list(theData)

        if radius < 0:
            radius = abs(radius)*fwhm
            
        if radius != 0:
            
            good_for_fit = less(sqrt((dataX-Xpos)**2+(dataY-Ypos)**2),radius)

            dataX   = compress(equal(good_for_fit,1),dataX)
            dataY   = compress(equal(good_for_fit,1),dataY)
            theData = compress(equal(good_for_fit,1),theData)

            # fUtilities.compress crash for array > 2100000
            # dataX,n   = fUtilities.compress(dataX,good_for_fit,1)
            # dataX     = dataX[:n]
            # dataY,n   = fUtilities.compress(dataY,good_for_fit,1)
            # dataY     = dataY[:n]
            # theData,n = fUtilities.compress(theData,good_for_fit,1)
            # theData   = theData[:n]
          
            
        # now use the data to fit a 2D-Gaussian
        
        try:
            self.PointingResult = fitBaseEllipticalGaussian(theData,dataX,dataY,\
                                                            fwhm=fwhm,\
                                                            gradient=gradient,\
                                                            circular=circular,\
                                                            Xpos=Xpos,Ypos=Ypos,\
                                                            fixedPos=fixedPos)
            self.showPointing(plot=plot,display=display,caption=caption,aspect=aspect)
        except BoaError, error:
            self.MessHand.warning('fit did not converge:')
            self.MessHand.warning(error.msg)
            self.PointingResult = -1

    def solvePointingOnMap(self,gradient=0, circular=0, radius=-10, \
                           Xpos = 0., Ypos = 0., fixedPos = 0, \
                           plot=0, display=1, caption='', aspect=1, style='idl4'):
        """
        DES: compute the offset on the data.Map object
        INP: (boolean)  gradient: shall we fit a gradient ? (default: no)
             (boolean)  circular: fit a cricular gaussian instead of an elliptical gaussian
             (float)    radius  : use only bolo inside this radiu
                                  (negative means multiple of beam - default: 10 beams)
             (float)  Xpos,Ypos : source position if using fixed position
             (boolean) fixedPos : if set, don't fit position, but use Xpos, Ypos
             (boolean)  plot    : do we plot the results? (default: no)
             (boolean)  display : display the result of the fit (default: yes)    
        OUT: store in self.PointingResult the results of the fit (i.e. all parameters
             as computed by mpfit routine). If mpfit failed, then self.PoitingResult
             is set to -1

             WARNING : No Smoothing should be applied to the map
             before using this function, or the fitted fwhm will be
             useless, use fine oversamp to make reasonable fit
        """

        if not self.Map.Data:
            self.MessHand.error('No map computed yet')
            return
        
        # fit a 2D-Gaussian on the map
        try:
            self.PointingResult = self.Map.extractSource(gradient=gradient,
                                                         circular=circular,
                                                         radius=radius,
                                                         Xpos=Xpos,Ypos=Ypos,
                                                         fixedPos=fixedPos)
            self.showPointing(plot=plot,display=display,caption=caption,
                              aspect=aspect,style=style)
        except BoaError, error:
            self.MessHand.warning('fit did not converge:')
            self.MessHand.warning(error.msg)
            self.PointingResult = -1

      
    def showPointing(self,plot=1,display=1,noMap=0,caption='',
                     aspect=1,style='idl4',limitsZ=[]):
        """
        DES: display results of last solvePointing (in text, and on the map if plot=1)
        INP: (logical) plot  : display the results on a map (default: no)
             (logical) display : display the result on screen (default: yes)
        """

        if not(self.PointingResult):
            self.MessHand.error('No pointing results to be displayed')
            return

        PointingResult = self.PointingResult

        if display:
            # Display out the result
            
            # Dictionnary to make links between keys in the pointing result
            # variable and the displayed name            
            pointingDict = {'gauss_x_offset': 'Delta Az ["]',\
                            'gauss_x_fwhm'  : 'FWHM_1   ["]',\
                            'gauss_y_offset': 'Delta El ["]',\
                            'gauss_y_fwhm'  : 'FWHM_2   ["]',\
                            'gauss_tilt'    : 'Tilt   [deg]',
                            'gauss_peak'    : 'Peak flux   '}
            
            if self.Map.WCS['CTYPE1'].find('RA') >= 0 and self.Map.WCS['CTYPE2'].find('DEC') >= 0:
                pointingDict['gauss_x_offset'] = 'RA     [deg]'
                pointingDict['gauss_y_offset'] = 'Dec    [deg]'
                # if map is equatorial, convert FWHM to arcsec for display
                PointingResult['gauss_x_fwhm']['value'] *= 3600.
                PointingResult['gauss_y_fwhm']['value'] *= 3600.
                
            # Select what will be displayed from the PointingResult
            toBeOutput = ['gauss_peak',\
                          'gauss_x_offset','gauss_y_offset',\
                          'gauss_x_fwhm','gauss_y_fwhm','gauss_tilt']
            
            outStr=""
            for item in toBeOutput:
                outStr+=pointingDict[item] + " = " + \
                         str(PointingResult[item]['value'])+'\n'
            self.MessHand.info(outStr)
            if self.Map.WCS['CTYPE1'].find('RA') >= 0 and self.Map.WCS['CTYPE2'].find('DEC') >= 0:
                # convert FWHM back to deg to compute the ellipsis
                PointingResult['gauss_x_fwhm']['value'] *= 1./3600.
                PointingResult['gauss_y_fwhm']['value'] *= 1./3600.
        
        if plot:
            # extract parameters of the ellipsis to plot
            xoff=PointingResult['gauss_x_offset']['value']
            yoff=PointingResult['gauss_y_offset']['value']
            xfwhm=PointingResult['gauss_x_fwhm']['value']
            yfwhm=PointingResult['gauss_y_fwhm']['value']
            tilt=PointingResult['gauss_tilt']['value']*pi/180.

            if not self.Map:
                self.doMap(style=style)
            else:
                if not noMap:
                    self.showMap(aspect=aspect,style=style,limitsZ=limitsZ) 
            # Overlay the pointing result
            Forms.ellipse(xoff,yoff,xfwhm,yfwhm,tilt)


    def arrayParameters(self,chanList=[], gradient=0, circular=0, radius=0, plot=0):
        """
        DES: determine the array parameters from the data
        INP: (i list) chanList : the channel list to be used (default: current list)
             (l)      gradient : remove a background gradient in the data (default: no)
             (l)      circular : fit a cricular gaussian instead of an elliptical gaussian
        """

        # Check the channel list
        chanList = self.BolometerArray.checkChanList(chanList)

        fwhm = self.BolometerArray.BeamSize
        
        arrayParamOffsets = []

        # Loop through all the channels
        for chan in chanList:
            if plot:
                self.doMap([chan],aspect=1)

            # By construction the source should be at 0,0
            self.solvePointing([chan],gradient=gradient,circular=circular,radius=radius,\
                               plot=plot, display=0)
            self._Map__Result = []
            
            arrayParamOffsets.append({'channel' : chan,\
                                      'result'  : self.PointingResult})

        self.arrayParamOffsets = arrayParamOffsets
        
    def updateArrayParameters(self,filename=None):
        """
        DES: Update the Parameters Offsets with the computed values
        INP: (str) filename : optional output file name
        """

        arrayParamOffsets = self.arrayParamOffsets
        Offsets = self.BolometerArray.Offsets
        refChan = self.BolometerArray.RefChannel
        gains   = self.FFCF_Gain
        
        # Find the reference offset
        refOffset = array([0,0])
        for iParam in arrayParamOffsets:
            if iParam['channel'] == refChan:
                refOffset = array([iParam['result']['gauss_x_offset']['value'],\
                                   iParam['result']['gauss_y_offset']['value']])

        if filename:
            fout = file(filename,'w')
            
        for iParam in arrayParamOffsets:
            chan = iParam['channel']
            Offsets[:,chan-1] = (Offsets[:,chan-1] - array([iParam['result']['gauss_x_offset']['value'],\
                                                            iParam['result']['gauss_y_offset']['value']]) \
                                 + refOffset).astype(Float32)
            oldgain = self.FF_Median[chan-1]
            gains[chan-1] = iParam['result']['gain']
            if filename:
                fout.write("%i %8.4f %8.4f %8.4f %8.4f\n"%(chan,gains[chan-1],oldgain,\
                                                           Offsets[0,chan-1],Offsets[1,chan-1]))
        self.BolometerArray.Offsets = Offsets
        self.FFCF_Gain = gains
        if filename:
            fout.close()
    
    #def reduce(self,datasetName='',obstoProc=[],radius=100.,update=0):
    def __plotSignalBefore(self,datasetName='',obstoProc=[],update=0):
        """
        DES: Process a Pointing scan - this method is called by the apexCalibrator
        INP: (string) datasetName: path to the dataset to be reduced
             (i list)   obstoProc: list of subscans to consider (default: all)
             (float)       radius: radius (in arcsec) to be used for fitting
        """
        self.read(inFile=datasetName,subscans=obstoProc,update=update)
        self.medianBaseline(subscan=0)

        # Automatic flagging of dead channels
        full = tolist_boa(self.BolometerArray.checkChanList([]))
        if full.count(self.BolometerArray.RefChannel):
            full.remove(self.BolometerArray.RefChannel)  # exclude ref. channel
        self.flagFractionRms(full)   # flag at median(rms)/10. and median*10.

        # We will produce 2 signal plots with 9 channels including the reference:
        # one before correlated noise removal, one after
        full = tolist_boa(self.BolometerArray.checkChanList([]))
        if not full.count(self.BolometerArray.RefChannel):
            self.unflagChannels(self.BolometerArray.RefChannel)
            full = tolist_boa(self.BolometerArray.checkChanList([]))
        numRef = full.index(self.BolometerArray.RefChannel)
        num1 = max(0,numRef-4)
        num2 = numRef+5
        if num2 > len(full):
            num2 = len(full)
        toPlot = full[num1:num2]
        self.signal(toPlot,caption="%i - raw data"%self.ScanParam.ScanNum)
        return toPlot
    
    def __plotSignalAfter(self,toPlot):
        full = tolist_boa(self.BolometerArray.checkChanList([]))
        # correlated noise removal: do not use Ref channel, may be affected by strong source
        # so, find the channel with lowest rms, excluding the dead ones
        full.remove(self.BolometerArray.RefChannel)  # exclude ref. channel
        rms = self.getChanListData('rms',full)
        mini, i_mini = rms[0], full[0]
        for i in range(1,len(full)):
            if rms[i] < mini:
                mini = rms[i]
                i_mini = full[i]
        full.append(self.BolometerArray.RefChannel)
        self.medianNoiseRemoval(full,chanRef=i_mini)
        self.medianBaseline(subscan=1)
        self.signal(toPlot,caption="%i - correlated noise subtracted"%self.ScanParam.ScanNum)

    def __reduceOnMap(self,radius=-2):
        # solve on map for a first guess
        self.doMap(oversamp=3.,noPlot=1)
        self.solvePointingOnMap(radius=radius,plot=1,display=0)
        # Results are now in self.PointingResult, (x,y) will be extracted to solve on data
        guess_X  = self.PointingResult['gauss_x_offset']['value']
        guess_Y  = self.PointingResult['gauss_y_offset']['value']
        self.MessHand.info(" 1st guess on map: (%5.1f, %5.1f)"%(guess_X,guess_Y))

    def __reduceOnData(self,guess_X,guess_Y):
        self.solvePointing(Xpos=guess_X,Ypos=guess_Y,radius=-2.,plot=1,display=1)
        # Results are now in self.PointingResult

    def writeModelData(self):
        """
        Generate one line to be written in the .dat file used for
        determining pointing model
        """
        # NB: this will work only if data have been read in with readAzEl0 = 1
        # and after the pointing has been reduced
        param = self.ScanParam
        # we need to find the 1st timestamp that is not flagged
        i0 = 0
        while self._flagHandler.isSetOnIndex(param.Flags, i0):
            i0 += 1
        sourceAz = param.AntAz0 - param.AzOff[i0]
        sourceEl = param.AntEl0 - param.ElOff[i0]
        encodAz  = param.EncAz0 - param.AzOff[i0]
        encodEl  = param.EncEl0 - param.ElOff[i0]
        PCA      = param.PDeltaCA
        PIE      = param.PDeltaIE
        FCA      = param.FDeltaCA
        FIE      = param.FDeltaIE
        scan     = param.ScanNum
        cCA      = self.PointingResult ['gauss_x_offset']['value'] / 3600.
        cIE      = self.PointingResult ['gauss_y_offset']['value'] / 3600.
        datLine = '%.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f %.6f %i\n' % \
                  (sourceAz, sourceEl, encodAz, encodEl,
                   cCA, cIE, PCA, PIE, FCA, FIE, scan)
        f = file('LABOCA-ABBA_1_2007-05-10.dat','a')
        f.write(datLine)
        f.close

        
