module f1
! """
! NAM: f1 (module f90) in file BoaF1.f90
! DES: module collecting f90 subroutines not (yet) belonging to a particular class:
!        gsmooth
!        reimage
!        compress
!        get_subscan_index
! USE: 
!      Call:  f90.f1.gsmooth(...)  etc. See examples in ftest.py
!      See BoaData.f90 for compilation instructions.
! """ 

!USE data  ! this links the data module which is available to all subroutines
contains
! !=============================================================================
! subroutine gsmooth(image,fwhm,ii,jj)
! ! """
! ! NAM: gsmooth (subroutine)
! ! DES: Smooth an image with a Gaussian kernel 
! !        f(x) = exp( -4 ln2 [x/FWHM]^2 )  where 4*ln2 = 2.772588
! !        f(FWHM/2) = 0.5,  f(FWHM) = 0.0625, f(1.5*FWHM)=0.002
! !      that extends to a*FWHM, where a=1.5 (may be changed). 
! !      The kernel is in fact a square box.
! !      For FWHM=1, it is of size 5x5, for 2 its 7x7, for 3 11x11 etc.
! !      The computation of the kernel values is approximate, simply 
! !      taking f(x) at the pixel center. 
! !      Blanked image pixels are ignored in the averaging. 
! !      Current blanking value = -1.e-7
! !      Blanks will be unblanked if they have nonblanked pixels in the 
! !      kernel range, else they stay at the blank value. Note that blanks
! !      are not considered to have zero value!
! ! INP: 
! !      image : input rank-2 array('f') with bounds (ii,jj); ii>1,jj>1
! !      fwhm  : input float. FWHM of Gaussian in pixel.
! ! OUT: 
! !      image : rank-2 array('f') with bounds (ii,jj). 
! !
! ! USE: image = f90.f1.gsmooth(image,fwhm)
! !      EXAMPLE:
! !        from Numeric import *
! !        import f90,time
! !        im = zeros(21,21,'f') +5.
! !        im[11,11] = 10.      # make a peak in the center
! !        im[11,10] = -1.e-7   # set to blanking value
! !        im = f90.f1.gsmooth(im,3.)
! ! """ 
! !-----------------------------------
! !INOUT
!  integer                          :: ii,jj
!  real                             :: fwhm
!  real, dimension(ii,jj)           :: image
!  !f2py intent(hide),depend(image) :: ii=shape(image,0),jj=shape(image,1)
!  !f2py intent(in,out)             :: image
!  !f2py intent(in)                 :: fwhm
! !-----------------------------------
! !LOCAL
!  real, parameter                  :: grange = 1.51, blank = -1.e7
!  real, allocatable,dimension(:,:) :: image_exp,kernel,subimage
!  logical, allocatable,dimension(:,:) :: subimage_mask
!  integer                          :: i,j,ng,ngg
! !-----------------------------------

!  if (fwhm .eq. 0) return

!  ng = nint(fwhm * grange) ! range of kernel in whole pixel
!  ngg = 2*ng+1             ! size of the kernel

!  ALLOCATE(image_exp(ii+2*ng,jj+2*ng))  ! make space for an expanded image
!  ALLOCATE(kernel(ngg,ngg),subimage(ngg,ngg),subimage_mask(ngg,ngg))

!  image_exp = blank    ! blank expanded image first
!  image_exp(ng+1:ii+ng,ng+1:jj+ng) = image ! then fill with image

!  do i=1,ngg        ! compute the radius^2 of each kernel pixel from kernel center
!     do j=1,ngg
!        kernel(i,j) = (ng+1-i)**2 + (ng+1-j)**2 
!     enddo
!  enddo

!  kernel = exp(-2.772588*kernel/fwhm**2)  ! that's the Kernel, center value = 1
! ! sum_kernel = sum(kernel) 

! ! do i=1,ngg      ! diagnostic output
! !   print "('kernel',20(f5.2))",kernel(i,:)
! ! enddo
! ! do i=1,ii
! !    print "(30(f4.0))",image(i,:)
! ! enddo

!  do i=ng+1,ng+ii       ! loop through all pixels in image
!     do j=ng+1,ng+jj
!        subimage = image_exp(i-ng:i+ng,j-ng:j+ng)    ! make subimage with size of kernel
!        subimage_mask = subimage.ne.blank            ! mask of non-blanked values
!        sum_kernel = sum(kernel,MASK=subimage_mask)  ! norm of masked kernel
!        if (any(subimage_mask)) &                    ! only if non-blanked values in kernel
!        image(i-ng,j-ng) = sum( kernel*subimage, MASK=subimage_mask)/sum_kernel ! smoothed pixel
!     enddo
!  enddo

! ! do i=1,ii  ! diagnostic output
! !    print "(30(f4.0))",image(i,:)
! ! enddo

!  DEALLOCATE(image_exp,kernel,subimage,subimage_mask)

! end subroutine gsmooth
!=============================================================================
subroutine reimage(image,data,xy,dxy,zero_offset,xy_scale,ii,jj,kk)
! """
! NAM: reimage (subroutine)
! DES: fill an image array with data in pixel boxes of given size
!        image = reimage(image,data,xy,dxy,zero_offset,xy_scale)
! INP: 
!        image       : input rank-2 array('f') with bounds (ii,jj)
!        data        : input rank-1 array('f') with bounds (kk)
!        xy          : input rank-2 array('f') with bounds (2,kk)
!        dxy         : input rank-2 array('f') with bounds (2,kk)  -- half width!
!        zero_offset : input rank-1 array('f') with bounds (2)
!                      bottom left corner user coordinate (eg arcsec) offset.
!        xy_scale    : input rank-1 array('f') with bounds (2)
!                      image scale, pixel per user coordinate (eg pix/arcsec)
! OUT:  
!        image       : rank-2 array('f') with bounds (ii,jj)
! USE: Example:
!        from Numeric import *
!        from fortran import f90
!        im = zeros((11,11),'f')  # odd size since coordinates refer to pix center
!        data    = array([10,20,30],'f')
!        xyscale = array([0.1,0.1],'f')
!        xy      = array([ [-50,0,20],[-30,0,20] ],'f')
!        dxy     = array([ [11,6,11] , [11,6,11] ],'f')
!        zerooffset = array([-50,-50],'f') # blc pix refers to -50,-50
!        im = f90.f1.reimage(im,data,xy,dxy,zerooffset,xyscale)
! """ 
!-----------------------------------
!INOUT
  integer                          :: ii,jj,kk
  real, dimension(ii,jj)           :: image
  real, dimension(kk),intent(in)   :: data
  real, dimension(2,kk)            :: xy,dxy
  real, dimension(2)               :: zero_offset,xy_scale
  !f2py intent(hide),depend(image) :: ii=shape(image,0),jj=shape(image,1)
  !f2py intent(hide),depend(data)  :: kk=shape(data)
  !f2py intent(in,out)             :: image
  !f2py intent(in)                 :: xy,dxy
!-----------------------------------
!LOCAL
  integer, dimension(2,kk) :: ixy_low,ixy_hig
!-----------------------------------
  image = 0.  ! erase image

  do i=1,2    ! rescale coordinates
     xy(i,:) = xy_scale(i) * ( xy(i,:) - zero_offset(i) ) ! box location
     dxy(i,:)= xy_scale(i) *  dxy(i,:) ! this is the half-width of the box
  enddo

  ixy_low = CEILING(xy - dxy)+1 ! lower and upper bounds of pixel box
  ixy_hig = FLOOR(xy + dxy)+1
!  ixy_hig(1,:) = MIN( ixy_hig(1,:), ii )
!  ixy_hig(2,:) = MIN( ixy_hig(2,:), jj )

  do k=1,kk   ! fill image
     print*,ixy_low(1,k),ixy_hig(1,k),ixy_low(2,k),ixy_hig(2,k)
     image(ixy_low(1,k):ixy_hig(1,k),ixy_low(2,k):ixy_hig(2,k))=data(k)
  enddo

end subroutine reimage
! !=============================================================================
! subroutine compress(xdata,xflag,flag_value,ii)
! ! """
! ! NAM: compress (subroutine)
! ! DES: compress array based on mask
! ! INP: data_input (f): 1-D array with data to be returned where flag = flag_value
! !      flag       (i): 1-D array, must be same size as data_input
! !      flag_value (i): select data where flag=flag_value
! ! OUT: tuple (x,ii): 1-D array of same size as data_input where first n elements are
! !                   those of data_input where flag = flag_value  
! ! USE: 
! !      Unfortunately it is not possible to pass from fortran to python an output
! !      array the size of which is computed in fortran. Therefore we must truncate
! !      the returned array at the size of the selected true mask elements.
! !
! !      Example:
! !            input_array = array(range(5),'f')
! !            flag_array  = array([0,1,0,1,0],'i')
! !            compress_array,nmax = f90.f1.compress(input_array,flag_array,1)
! !            compress_array = compress_array[0:nmax]
! !            print "nmax compressed:",nmax,compress_array
! ! """ 
! !-----------------------------------
! !INOUT
!   integer                :: ii
!   real,    dimension(ii) :: xdata
!   integer, dimension(ii) :: xflag
!   integer, intent(in)    :: flag_value
!   !f2py intent(in,out,copy),depend(xflag)    :: ii=shape(xflag)
!   !f2py intent(in,out,copy)                  :: xdata
! !-----------------------------------
! !LOCAL
!   logical, dimension(ii) :: flag_mask   
!   integer :: mask_true
!   integer :: data_size,flag_size
! !-----------------------------------
!     flag_size = size(xflag)
!     data_size = size(xdata)
!     if (data_size.ne.flag_size) &       !   Need to also add an exit in this case
!         print*,"E: data and index have different size:",data_size,flag_size

!     flag_mask  = xflag.eq.flag_value    ! make a mask
!     mask_true  = COUNT(flag_mask)       ! counts the number of TRUE in mask

!     if(mask_true>0) then                ! only if there are at least 2 unmasked data points
!     xdata = PACK(xdata,MASK=flag_mask)  ! this does not overwrite xdata on the py level
!                                         ! because i specify intent(copy)
!     else
!        print*,"W: no data points selected"
!     end if

!     ii = mask_true    ! return mask_value to py without creating an extra input argument

! end subroutine compress

! !=============================================================================
! subroutine get_subscan_index(xsubscan_index,ii)
! ! """
! ! NAM: get_subscan_index (subroutine)
! ! DES: get the beginning index of each subscan data in phase array
! !      by comparing the subscan begin time with lst_p.
! !      Result must be fetched from f90.data module.
! !      Might be good to specify the ending index as well if we sample 
! !      data in between subscans.
! ! INP: data.subscan_time   f
! !      data.lst_p          f
! ! OUT: data.subscan_index  i
! ! """ 
! !-----------------------------------
! !INOUT
!   integer,dimension(2,ii)         :: xsubscan_index
!   integer                         :: ii 
!   !f2py intent(hide),depend(subscan_time):: ii=shape(subscan_time)
!   !f2py intent(in,out)            :: xsubscan_index
! !-----------------------------------
! !LOCAL
!   integer :: i
! !-----------------------------------
!  do i=1,size(subscan_time)
!     xsubscan_index(1,i) = SUM(MINLOC(lst_p(:,1),MASK=lst_p(:,1).ge.subscan_time(i)))

! ! is there a more elegent way to turn a 1-element array into a scalar than SUM?

! if(i.gt.1.and.xsubscan_index(1,i).le.xsubscan_index(1,i-1)+1) & ! check there is more than 1
!          print*,"E: not enough data in subscan",i               ! data point in each subscan

!  end do

!   xsubscan_index(2,:) = cshift(xsubscan_index(1,:),1)
!   xsubscan_index(2,size(subscan_time)) = size(lst_p(:,1))+1

! end subroutine get_subscan_index
!============================================================================
END MODULE f1
!============================================================================
