module baseline
! """
! NAM: baseline (module f90) in file BoaFortranBaseLine.f90
! DES: module collecting f90 subroutines:
!      polyfit
!      addpoly
! USE: Call:  f90.baseline.polyfit(...)  etc. See examples in ftest.py
! """ 

contains
!=============================================================================
subroutine polyfit(ydata,xdata,flag,poly,mean,rms,nx,deg)
! """
! NAM: polyfit (f90 subroutine)
! DES: polynomial fit to data with singular value decomposition
!      using modified routines from from Numerical Recipes.
!      Produces polynomial coefficients and fit array matching the data (including
!      the flagged!), but does not subtract fit from data. 
!      Fit may be plotted easily before subtraction.
!      Before fitting, the data are rescaled to a range approx. -1 to 1
!      for better stability. This may cost some time though.
! INP: 
!      ydata        = f 1D array: Y data to be fitted
!      xdata        = f 1D array: X values - result: Y = Pn(X) (length nx)
!      flag         = i 1D array: flag associated with ydata (length nx)
!      deg          = i polynom order + 1
!      poly         = f 1D array of size order of fit, content irrelevant (length deg)
!      mean         = f mean of data. Must be known!
!      rms          = f rms of data. Must be known!
!
! OUT: tuple: Fit,Poly
!      Fit  = f 1D ydata array where data replaced by poly fit
!      Poly = f 1D polynomial coefficients
!
! USE: ...
! """
  USE svd
!INOUT ----------------------------
  integer                  :: nx,deg ! these must appear in the call parameter list but are hidden
  real,    dimension(nx)   :: ydata
  real,    dimension(nx)   :: xdata
  integer, dimension(nx)   :: flag
  real,    dimension(deg)  :: poly
  real                     :: mean,rms
  !f2py intent(hide),depend(xdata) :: nx=shape(xdata,0)
  !f2py intent(hide),depend(poly)  :: deg=shape(poly,0)
  !f2py intent(in,out)             :: poly
  !f2py intent(in,out)             :: ydata
!LOCAL ---------------------------- 
  integer                        :: i
  real, parameter                :: spread=0.1
  real                           :: chisq
  real, dimension(deg)           :: w
  real, dimension(deg,deg)       :: cvm,v
  logical, dimension(nx)         :: flag_mask
  real, dimension(nx)            :: xall,yall
  real,allocatable,dimension(:)  :: x,y,sig    ! their dim depend on mask
! ---------------------------------

  flag_mask = flag.eq.0          ! make a mask
  mask_count = COUNT(flag_mask)  ! number of TRUE in mask

  if(mask_count>1) then          ! only if there are at least 2 unmasked data points

     ALLOCATE(x(mask_count))
     ALLOCATE(y(mask_count))
     ALLOCATE(sig(mask_count))
     
     y = PACK(ydata,MASK=flag_mask)
     y = (y - mean) / rms

     x = PACK(xdata,MASK=flag_mask) 
     x_min   = minval(x)
     x_scale = maxval(x)-x_min
     x = 2.*(x - x_min)/x_scale - 1.
     sig = y*0. + SPREAD*rms

     call svdfit(x,y,sig,poly,v,w,chisq)     ! fit order from size of poly,v,w

     !  write(*,'("I: fit=",2x,10(ES11.3))') poly
     !  write(*,'("I: rms=",ES11.3)') sqrt(chisq/mask_count)

     ! Now write the fit out. Do not subtract fit from data.
     xall = 2*(xdata - x_min)/x_scale-1
     yall = xall*0. + poly(1)
     DO i=2,deg
        yall = yall + poly(i)*xall**(i-1)
     END DO

     ydata = yall * rms + mean
     ! this will not overwrite the data on the py level, it's just to produce the output.

     DEALLOCATE(x,y,sig)
 end if

end subroutine polyfit

!=============================================================================
subroutine addpoly(xdata,poly,cha_mean,cha_rms,n_phase,n_channel,ii,jj,kk,ll)

! """
! NAM: addpoly (f90 subroutine)
! DES: add polynomial to data 
! """
!INOUT ----------------------------
  real,    dimension(ii,jj,kk)     :: xdata
  real,    dimension(jj,kk)        :: cha_mean,cha_rms
  integer, intent(in)              :: n_phase,n_channel
  real,    dimension(ll),intent(in):: poly
  integer :: ii,jj,kk,ll 
  !f2py intent(hide),depend(xdata) :: ii=shape(xdata,0),jj=shape(xdata,1),kk=shape(xdata,2)
  !f2py intent(hide),depend(poly)  :: ll=shape(poly)
  !f2py intent(in,out)             :: xdata
!LOCAL ---------------------------- 
  integer                         :: n_ph, n_ch
  logical, dimension(ii)          :: flag_mask
  real,    dimension(ii)          :: x,y
! ---------------------------------

  n_ph = n_phase + 1     ! conform to python convention  0,1,2
  n_ch = n_channel + 1

  y = (xdata(:,n_ph,n_ch) - cha_mean(n_ph,n_ch))/cha_rms(n_ph,n_ch)
  !x = lst_p(:,n_ph)
  x_min   = minval(x)
  x_scale = maxval(x)-x_min
  x       = 2*(x - x_min)/x_scale-1

  y = y + poly(1)
  DO i=2,ll
     y = y + poly(i)*x**(i-1)
  END DO

  xdata(:,n_ph,n_ch) = y * cha_rms(n_ph,n_ch) + cha_mean(n_ph,n_ch)

end subroutine addpoly
!============================================================================
end module baseline
!============================================================================
