module snf
! """
! NAM: snf (module f90) in file BoaSNF.f90
! DES: module collecting f90 subroutines for sky noise subtraction:
!        corfit
!        cmatrix
! USE: f90.snf.corfit(...)  etc. See examples in TestFB.py
!      For compilation instruction see BoaData.f90
! """ 

contains
!=============================================================================
subroutine corfit(xdata,xnoise,xflag,n_reference,slope,nx,ny)
! """
! NAM: corfit (f90 subroutine)
! DES: order-1 (linear) fit to data vs. noise
!      Produces fit coefficients and fit array matching the data (including
!      the flagged!).
! INP: xdata        = data array: currData.DataRed
!      xflag        = flag array: currData.DataFlags
!      xnoise       = cor noise array: currData.DataNoi
!        dimensions: nx (nb. integ.) x ny (nb. channels)
!      n_reference  = reference channel number 
!
! OUT: xdata        = data array, replaced with fit
!      slope        = array with slopes of linear fit
! """

  USE svd

!INOUT ----------------------------
  real,    dimension(nx,ny)     :: xdata,xnoise
  real,    dimension(ny)        :: a,slope
  integer, dimension(nx,ny)     :: xflag
  integer, intent(in)           :: n_reference
  integer :: nx,ny 
  !f2py intent(hide),depend(xdata) :: nx=shape(xdata,0),ny=shape(xdata,1)
  !f2py intent(in,out)             :: xdata
  !f2py intent(in)                 :: xnoise
  !f2py intent(out,hide)           :: slope
!LOCAL ---------------------------- 
  integer                        :: n_ch, n_ref, i
  real                           :: chisq,x_max,q,aa,bb
  logical, dimension(nx)         :: flag_mask
  real, dimension(nx)            :: xall
  real,allocatable,dimension(:)  :: x,y    ! its dim depends on mask
! ---------------------------------

!call openmes   ! open the current message file

n_ref = n_reference + 1   ! conform to python convention  0,1,2

DO n_ch=1,ny        ! loop through all channels
  if(n_reference .lt. 0) then
     n_ref = n_ch  ! in this case correlate noise not data
  endif

  flag_mask = (xflag(:,n_ref).eq.0).and.(xflag(:,n_ch).eq.0)  ! make a mask
  mask_count = COUNT(flag_mask)        ! number of TRUE in mask

  if(mask_count>1) then       ! only if there are at least 2 unmasked data points
     ALLOCATE(x(mask_count))
     ALLOCATE(y(mask_count))

     y = PACK(xdata(:,n_ch),MASK=flag_mask)
     x = PACK(xnoise(:,n_ref),MASK=flag_mask) 
     xall = xnoise(:,n_ref)

     call fit(x,y,aa,bb,chisq)
     a(n_ch)=aa
     slope(n_ch)=bb

     ! Now write the fit out.
     if((n_ch .ne. n_ref).or.(n_reference .lt. 0)) then
        xdata(:,n_ch) =  a(n_ch) + slope(n_ch)*xall
     endif
     ! this will not overwrite the data on the py level, it's just to produce the output.

     !--- write diagnostics to log file
     !write(11,'(" I: ch=",i3," fit=",2(F7.3)," rms=",ES11.3,I6)') &
     !     n_ch,a(n_ch),slope(n_ch),sqrt(chisq/mask_count),mask_count

     DEALLOCATE(x,y)
  end if
END DO

!close(11) ! do not forget to close log file again

end subroutine corfit
!=============================================================================
subroutine cmatrix(xdata,xflag,ccmatrix,nx,ny)
! """
! NAM: cmatrix  (sub)
! DES: compute correlation matrix
! INP: xdata     = entire data array: currData.DataRed
!      xflag     = entire flag array: currData.DataFlags
!        dimensions: nx (nb. integ.) x ny (nb. channels)
! OUT: ccmatrix (channel,channel) = correlation matrix
! """
!-----------------------------------
!INOUT
  real,    dimension(nx,ny)     :: xdata
  integer, dimension(nx,ny)     :: xflag
  real,    dimension(ny,ny)     :: ccmatrix
  integer                       :: nx,ny
  !f2py intent(hide),depend(xflag) :: nx=shape(xflag,0),ny=shape(xflag,1)
  !f2py intent(in,out)             :: ccmatrix
  !f2py intent(in)                 :: xdata,xflag
!-----------------------------------
!LOCAL
  integer                        :: nc1,nc2,mask_count,imax
  logical,dimension(nx)          :: flag_mask
  real,allocatable,dimension(:)  :: arr1,arr2
!-----------------------------------

DO nc1=1,ny        ! loop through all channels
   DO nc2=nc1,ny   ! loop through all channels

      flag_mask  = (xflag(:,nc1).eq.0).and.(xflag(:,nc2).eq.0) ! make a mask
      mask_count = COUNT(flag_mask)            ! number of TRUE in mask

      if(mask_count>1) then       ! only if there are at least 2 numbers           
         ALLOCATE(arr1(mask_count),arr2(mask_count))
         arr1 = PACK(xdata(:,nc1),MASK=flag_mask)
         arr2 = PACK(xdata(:,nc2),MASK=flag_mask)

         cc =  abs(sum(arr1*arr2) / sqrt(sum(arr1*arr1) * sum(arr2*arr2)))
         ! cc by definition positive

         ccmatrix(nc1,nc2) = cc
         ccmatrix(nc2,nc1) = cc
         DEALLOCATE(arr1,arr2)
      endif
   END DO
END DO

!call openmes   ! open the current message file
!imax = min(14,ny)
!write(11,'(" I: f90 correlation matrix computed")')
!write(11,'(4x,14(i7))') (i,i=1,imax)
!do nc1=1,imax
!   write(11,'(i3,1x,14(f7.3))') nc1,ccmatrix(nc1,1:imax)
!enddo
!close(11) ! do not forget to close it again

return
end subroutine cmatrix
!============================================================================
end module snf
!============================================================================
